''' For Standalone biothing tracking '''
import sys, os, base64, datetime, hashlib, hmac, json, logging 
from tornado.httpclient import HTTPRequest, AsyncHTTPClient

# Key derivation functions. See:
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
def sign(key, msg):
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

def getSignatureKey(key, date_stamp, regionName, serviceName):
    kDate = sign(('AWS4' + key).encode('utf-8'), date_stamp)
    kRegion = sign(kDate, regionName)
    kService = sign(kRegion, serviceName)
    kSigning = sign(kService, 'aws4_request')
    return kSigning

def tracking_callback(response):
    logging.debug("GA response: {}".format(str(response)))
    logging.debug("Body: {}".format(response.buffer.getvalue()))
    logging.debug("Response Headers: {}".format(str(response.headers)))
    logging.debug("Request Headers: {}".format(str(response.request.headers)))
    logging.debug("Request Body: {}".format(str(response.request.body)))
    return

# This is a mixin for biothing handlers, and references class variables from that class, cannot be used
# without mixing in
class StandaloneTrackingMixin:
    def self_track(self, data={}):
        no_tracking = self.get_argument('no_tracking', None)
        access_key = self.web_settings.STANDALONE_AWS_CREDENTIALS.get('AWS_ACCESS_KEY_ID', False)
        secret_key = self.web_settings.STANDALONE_AWS_CREDENTIALS.get('AWS_SECRET_ACCESS_KEY', False)
        if not no_tracking and self.web_settings.STANDALONE_TRACKING_URL and access_key and secret_key:
            self.web_settings.tracking_payload.append(json.dumps({
                "action": data.get('action', 'NA'),
                "biothing": self.web_settings.ES_DOC_TYPE,
                "category": data.get('category', 'NA')
            }))
            logging.debug("tracking_payload size: {}".format(len(self.web_settings.tracking_payload)))
            if (len(self.web_settings.tracking_payload) == self.web_settings.STANDALONE_TRACKING_BATCH_SIZE):

                # ************* REQUEST VALUES *************
                request_body = '\n'.join(self.web_settings.tracking_payload)
                # logging.debug("Standalone Request Body: {}".format(request_body))
                # reset payload
                self.web_settings.tracking_payload = []
                method = 'POST'
                service = 'execute-api'
                endpoint = self.web_settings.STANDALONE_TRACKING_URL
                host = endpoint.split('://')[1].split('/')[0]
                canonical_uri = endpoint.split(host)[1]
                region = 'us-west-1'

                # POST requests use a content type header.
                content_type = 'application/x-amz-json-1.0'
                content_length = len(request_body)

                # Create a date for headers and the credential string
                t = datetime.datetime.utcnow()
                amz_date = t.strftime('%Y%m%dT%H%M%SZ')
                date_stamp = t.strftime('%Y%m%d') # Date w/o time, used in credential scope

                # ************* TASK 1: CREATE A CANONICAL REQUEST *************
                # http://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html

                # Step 1 is to define the verb (GET, POST, etc.)--already done.

                # Step 2: Create canonical URI--the part of the URI from domain to query 
                # string (use '/' if no path) -- already done.

                ## Step 3: Create the canonical query string. In this example, request
                # parameters are passed in the body of the request and the query string
                # is blank.
                canonical_querystring = ''

                # Step 4: Create the canonical headers. Header names must be trimmed
                # and lowercase, and sorted in code point order from low to high.
                # Note that there is a trailing \n.
                canonical_headers = 'content-length:' + '{}'.format(content_length) + '\n' + 'content-type:' + content_type + '\n' + 'host:' + host + '\n' + 'x-amz-date:' + amz_date + '\n'

                # Step 5: Create the list of signed headers. This lists the headers
                # in the canonical_headers list, delimited with ";" and in alpha order.
                # Note: The request can include any headers; canonical_headers and
                # signed_headers include those that you want to be included in the
                # hash of the request. "Host" and "x-amz-date" are always required.
                signed_headers = 'content-length;content-type;host;x-amz-date'

                # Step 6: Create payload hash. 
                payload_hash = hashlib.sha256(request_body.encode('utf-8')).hexdigest()

                # Step 7: Combine elements to create create canonical request
                canonical_request = method + '\n' + canonical_uri + '\n' + canonical_querystring + '\n' + canonical_headers + '\n' + signed_headers + '\n' + payload_hash


                # ************* TASK 2: CREATE THE STRING TO SIGN*************
                # Match the algorithm to the hashing algorithm you use, either SHA-1 or
                # SHA-256 (recommended)
                algorithm = 'AWS4-HMAC-SHA256'
                credential_scope = date_stamp + '/' + region + '/' + service + '/' + 'aws4_request'
                string_to_sign = algorithm + '\n' +  amz_date + '\n' +  credential_scope + '\n' +  hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()


                # ************* TASK 3: CALCULATE THE SIGNATURE *************
                # Create the signing key using the function defined above.
                signing_key = getSignatureKey(secret_key, date_stamp, region, service)

                # Sign the string_to_sign using the signing_key
                signature = hmac.new(signing_key, (string_to_sign).encode('utf-8'), hashlib.sha256).hexdigest()


                # ************* TASK 4: ADD SIGNING INFORMATION TO THE REQUEST *************
                # Put the signature information in a header named Authorization.
                authorization_header = algorithm + ' ' + 'Credential=' + access_key + '/' + credential_scope + ', ' +  'SignedHeaders=' + signed_headers + ', ' + 'Signature=' + signature

                req = HTTPRequest(url=self.web_settings.STANDALONE_TRACKING_URL, method=method, body=request_body, 
                    headers={
                        "Content-Type": content_type,
                        "Content-Length": content_length,
                        "X-Amz-Date": amz_date,
                        "Authorization": authorization_header,
                        "Host": host
                    }
                )

                #now send actual async requests
                http_client = AsyncHTTPClient()
                http_client.fetch(req)#, callback=tracking_callback)



'''
NOTE: This is deprecated!

Utils for running parallel jobs on IPython cluster.
'''
import os, sys
import time
import copy
from ipyparallel import Client

# ipcluster expecting to be in the folder where the app config is
# because ipcluster will just import this file/module, kind of as if
# it was building a main script from this lib
import config, biothings
biothings.config_for_app(config)
from biothings.utils.common import timesofar, ask


def run_jobs_on_ipythoncluster(worker, task_list, shutdown_ipengines_after_done=False):

    t0 = time.time()
    rc = Client(config.CLUSTER_CLIENT_JSON)
    lview = rc.load_balanced_view()
    cnt_nodes = len(lview.targets or rc.ids)
    print("\t# nodes in use: {}".format(cnt_nodes))
    lview.block = False
    # move to app path
    lview.map(os.chdir, [config.APP_PATH] * cnt_nodes)
    print("\t# of tasks: {}".format(len(task_list)))
    print("\tsubmitting...", end='')
    job = lview.map_async(worker,task_list)
    print("done.")
    try:
        job.wait_interactive()
    except KeyboardInterrupt:
        #handle "Ctrl-C"
        if ask("\nAbort all submitted jobs?") == 'Y':
            lview.abort()
            print("Aborted, all submitted jobs are cancelled.")
        else:
            print("Aborted, but your jobs are still running on the cluster.")
        return

    if len(job.result()) != len(task_list):
        print("WARNING:\t# of results returned ({}) != # of tasks ({}).".format(len(job.result()), len(task_list)))
    print("\ttotal time: {}".format(timesofar(t0)))

    if shutdown_ipengines_after_done:
        print("\tshuting down all ipengine nodes...", end='')
        lview.shutdown()
        print('Done.')
    return job.result()


def collection_partition(src_collection_list, step=100000):
    if not isinstance(src_collection_list, (list, tuple)):
        src_collection_list = [src_collection_list]

    kwargs = {}
    kwargs['limit'] = step
    for src_collection in src_collection_list:
        _kwargs = copy.copy(kwargs)
        _kwargs['src_collection'] = src_collection.name
        _kwargs['src_db'] = src_collection.database.name
        _kwargs['server'] = src_collection.database.connection.host
        _kwargs['port'] = src_collection.database.connection.port

        cnt = src_collection.count()
        for s in range(0, cnt, step):
            __kwargs = copy.copy(_kwargs)
            __kwargs['skip'] = s
            yield __kwargs



import os
import asyncio
import json
from functools import partial

from biothings import config as btconfig
import biothings.hub.dataload.uploader as uploader
from biothings.utils.backend import DocESBackend
from biothings.utils.es import IndexerException


class BiothingsUploader(uploader.BaseSourceUploader):

    name = None

    # Specify the backend this uploader should work with. Must be defined before instantiation
    # (can be an instance or a partial() returning an instance)
    TARGET_BACKEND = None

    # Specify the syncer function this uploader will use to apply diff
    # (can be an instance or a partial() returning an instance)
    SYNCER_FUNC = None

    # should we delete index before restoring snapshot if index already exist ?
    AUTO_PURGE_INDEX = False

    def __init__(self, *args, **kwargs):
        super(BiothingsUploader, self).__init__(*args, **kwargs)
        self._target_backend = None
        self._syncer_func = None

    @property
    def target_backend(self):
        if not self._target_backend:
            if type(self.__class__.TARGET_BACKEND) == partial:
                self._target_backend = self.__class__.TARGET_BACKEND()
            else:
                self._target_backend = self.__class__.TARGET_BACKEND
            assert type(
                self._target_backend
            ) == DocESBackend, "Only ElasticSearch backend is supported (got %s)" % type(
                self._target_backend)
        return self._target_backend

    @property
    def syncer_func(self):
        if not self._syncer_func:
            self._syncer_func = self.__class__.SYNCER_FUNC
        return self._syncer_func

    @asyncio.coroutine
    def load(self, *args, **kwargs):
        return super().load(steps=["data"], *args, **kwargs)

    @asyncio.coroutine
    def update_data(self, batch_size, job_manager):
        """
        Look in data_folder and either restore a snapshot to ES
        or apply diff to current ES index
        """
        # determine if it's about a snapshot/full and diff/incremental
        # we should have a json metadata matching the release
        self.prepare_src_dump()  # load infor from src_dump
        release = self.src_doc.get("download", {}).get("release")
        assert release, "Can't find release information in src_dump document"
        build_meta = json.load(
            open(os.path.join(self.data_folder, "%s.json" % release)))
        if build_meta["type"] == "full":
            res = yield from self.restore_snapshot(build_meta,
                                                   job_manager=job_manager)
        elif build_meta["type"] == "incremental":
            res = yield from self.apply_diff(build_meta,
                                             job_manager=job_manager)
        return res

    def get_snapshot_repository_config(self, build_meta):
        """Return (name,config) tuple from build_meta, where
        name is the repo name, and config is the repo config"""
        repo_name, repo_settings = list(
            build_meta["metadata"]["repository"].items())[0]
        return (repo_name, repo_settings)

    @asyncio.coroutine
    def restore_snapshot(self, build_meta, job_manager, **kwargs):
        idxr = self.target_backend.target_esidxer
        repo_name, repo_settings = self.get_snapshot_repository_config(
            build_meta)
        # first check if snapshot repo exists
        # do we need to enrich with some credentials ? (there are part of repo creation JSON settings)
        if repo_settings.get(
                "type") == "s3" and btconfig.STANDALONE_AWS_CREDENTIALS.get(
                    "AWS_ACCESS_KEY_ID"):
            repo_settings["settings"][
                "access_key"] = btconfig.STANDALONE_AWS_CREDENTIALS[
                    "AWS_ACCESS_KEY_ID"]
            repo_settings["settings"][
                "secret_key"] = btconfig.STANDALONE_AWS_CREDENTIALS[
                    "AWS_SECRET_ACCESS_KEY"]
            repo_settings["settings"]["readonly"] = True
        try:
            repo = idxr.get_repository(repo_name)
            # ok it exists, check if settings are the same
            if repo[repo_name] != repo_settings:
                # different, raise exception so it's handles in the except
                self.logger.info(
                    "Repository '%s' was found but settings are different, it needs to be created again"
                    % repo_name)
                self.logger.debug("Existing setting: %s" % repo[repo_name])
                self.logger.debug("Required (new) setting: %s" % repo_settings)
                raise IndexerException
        except IndexerException:
            # ok, it doesn't exist let's try to create it
            try:
                repo = idxr.create_repository(repo_name, repo_settings)
            except IndexerException as e:
                if repo_settings["settings"].get("url"):
                    raise uploader.ResourceError("Could not create snapshot repository. Check elasticsearch.yml configuration "
                                                 + "file, you should have a line like this: "
                                                 + 'repositories.url.allowed_urls: "%s*" ' % repo_settings["settings"]["url"]
                                                 + "allowing snapshot to be restored from this URL. Error was: %s" % e)
                else:
                    # try to create repo without key/secret, assuming it's already configured in ES keystore
                    if repo_settings["settings"].get("access_key"):
                        repo_settings["settings"].pop("access_key")
                        repo_settings["settings"].pop("secret_key")
                        try:
                            repo = idxr.create_repository(
                                repo_name, repo_settings)
                        except IndexerException as e:
                            raise uploader.ResourceError("Could not create snapshot repository, even assuming "
                                                         + "credentials configured in keystore: %s" % e)
                    else:
                        raise uploader.ResourceError(
                            "Could not create snapshot repository: %s" % e)

        # repository is now ready, let's trigger the restore
        snapshot_name = build_meta["metadata"]["snapshot_name"]
        pinfo = self.get_pinfo()
        pinfo["step"] = "restore"
        pinfo["description"] = snapshot_name

        def get_status_info():
            try:
                res = idxr.get_restore_status(idxr._index)
                return res
            except Exception as e:
                # somethng went wrong, report as failure
                return {"status": "FAILED %s" % e}

        def restore_launched(f):
            try:
                self.logger.info("Restore launched: %s" % f.result())
            except Exception as e:
                self.logger.error("Error while lauching restore: %s" % e)
                raise e

        self.logger.info("Restoring snapshot '%s' to index '%s' on host '%s'" %
                         (snapshot_name, idxr._index, idxr.es_host))
        job = yield from job_manager.defer_to_thread(
            pinfo,
            partial(idxr.restore,
                    repo_name,
                    snapshot_name,
                    idxr._index,
                    purge=self.__class__.AUTO_PURGE_INDEX))
        job.add_done_callback(restore_launched)
        yield from job
        while True:
            status_info = get_status_info()
            status = status_info["status"]
            self.logger.info("Recovery status for index '%s': %s" %
                             (idxr._index, status_info))
            if status in ["INIT", "IN_PROGRESS"]:
                yield from asyncio.sleep(
                    getattr(btconfig, "MONITOR_SNAPSHOT_DELAY", 60))
            else:
                if status == "DONE":
                    self.logger.info("Snapshot '%s' successfully restored to index '%s' (host: '%s')" %
                                     (snapshot_name, idxr._index, idxr.es_host), extra={"notify": True})
                else:
                    e = uploader.ResourceError("Failed to restore snapshot '%s' on index '%s', status: %s" %
                                               (snapshot_name, idxr._index, status))
                    self.logger.error(e)
                    raise e
                break
        # return current number of docs in index
        return self.target_backend.count()

    @asyncio.coroutine
    def apply_diff(self, build_meta, job_manager, **kwargs):
        self.logger.info("Applying incremental update from diff folder: %s" %
                         self.data_folder)
        meta = json.load(open(os.path.join(self.data_folder, "metadata.json")))
        # old: index we want to update
        old = (self.target_backend.target_esidxer.es_host,
               self.target_backend.target_name,
               self.target_backend.target_esidxer._doc_type)
        # new: index's data we will reach once updated (just informative)
        new = (self.target_backend.target_esidxer.es_host,
               meta["new"]["backend"],
               self.target_backend.target_esidxer._doc_type)
        yield from self.syncer_func(old_db_col_names=old,
                                    new_db_col_names=new,
                                    diff_folder=self.data_folder)
        # return current number of docs in index (even if diff update)
        return self.target_backend.count()

    def clean_archived_collections(self):
        pass



"""
DataTransform Histogram class - track keylookup statistics
"""


class Histogram(object):
    """
    Histogram - track keylookup statistics
    """

    def __init__(self):
        self.io_histogram = {}
        self.edge_histogram = {}

    def __str__(self):
        res = {
            'io_report': self.io_histogram,
            'edge_report': self.edge_histogram
        }
        return str(res)

    def update_edge(self, vert1, vert2, size):
        """
        Update the edge histogram
        """
        key = self._construct_key(vert1, vert2)
        self._increment(self.edge_histogram, key, size)

    def update_io(self, input_type, output_type, size):
        """
        Update the edge histogram
        """
        key = self._construct_key(input_type, output_type)
        self._increment(self.io_histogram, key, size)

    @staticmethod
    def _construct_key(obj1, obj2):
        return "{}-->{}".format(obj1, obj2)

    @staticmethod
    def _increment(hist, key, size):
        if size > 0:
            if key not in hist.keys():
                hist[key] = int(size)
            else:
                hist[key] += int(size)



import asyncio
from tornado.web import RequestHandler
import logging
import datetime
import pandas.io.json as pdjson

from biothings import config


class DefaultHandler(RequestHandler):
    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Content-Type', 'application/json')
        # part of pre-flight requests
        self.set_header('Access-Control-Allow-Methods',
                        'PUT, DELETE, POST, GET, OPTIONS')
        self.set_header(
            'Access-Control-Allow-Headers',
            'Content-Type,X-BioThings-API,X-Biothings-Access-Token')

    def write(self, result):
        super(DefaultHandler, self).write(
            pdjson.dumps({
                "result": result,
                "status": "ok"
            }, iso_dates=True))

    def write_error(self, status_code, **kwargs):
        self.set_status(status_code)
        super(DefaultHandler, self).write({
            "error":
            str(kwargs.get("exc_info", [None, None, None])[1]),
            "status":
            "error",
            "code":
            status_code
        })

    # defined by default so we accept OPTIONS pre-flight requests
    def options(self, *args, **kwargs):
        logging.debug("OPTIONS args: %s, kwargs: %s" % (args, kwargs))


class BaseHandler(DefaultHandler):
    def initialize(self, managers, **kwargs):
        self.managers = managers


class GenericHandler(DefaultHandler):
    def initialize(self, shell, **kwargs):
        self.shell = shell

    def get(self, *args, **kwargs):
        logging.debug("GET args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method GET not allowed", None))

    def post(self, *args, **kwargs):
        logging.debug("POST args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method POST not allowed", None))

    def put(self, *args, **kwargs):
        logging.debug("PUT args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method PUT not allowed", None))

    def delete(self, *args, **kwargs):
        logging.debug("DELETE args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405,
                         exc_info=(None, "Method DELETE not allowed", None))

    def head(self, *args, **kwargs):
        logging.debug("HEAD args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method HEAD not allowed", None))


class RootHandler(DefaultHandler):
    def initialize(self, features, hub_name=None, **kwargs):
        self.features = features
        self.hub_name = hub_name

    @asyncio.coroutine
    def get(self):
        self.write({
            "name": self.hub_name or \
            getattr(config, "HUB_NAME", None),
            "biothings_version":
            getattr(config, "BIOTHINGS_VERSION", None),
            "app_version":
            getattr(config, "APP_VERSION", None),
            "icon":
            getattr(config, "HUB_ICON", None),
            "now":
            datetime.datetime.now(),
            "features":
            self.features,
        })



from .common import is_str


def last_element(d, key_list):
    """Return the last element and key for a document d given a docstring.

    A document d is passed with a list of keys key_list.  A generator is then
    returned for all elements that match all keys.  Not that there may be 
    a 1-to-many relationship between keys and elements due to lists in the document.

    :param d: document d to return elements from
    :param key_list: list of keys that specify elements in the document d
    :return: generator for elements that match all keys
    """
    # preconditions
    if not d or not key_list:
        return
    k = key_list.pop(0)
    # termination
    if not key_list:
        yield k, d
    # recursion
    else:
        try:
            t = d[k]
        except KeyError:
            return # key does not exist
        except TypeError:
            return # not sub-scriptable
        if isinstance(t, dict):                
            yield from last_element(t, key_list)
        elif isinstance(t, list):
            for l in t:
                yield from last_element(l, key_list.copy())
        elif isinstance(t, tuple):
            # unsupported type
            raise ValueError("unsupported type in key {}".format(k))
    

def key_value(dictionary, key):
    """Return a generator for all values in a dictionary specific by a dotstirng (key)

    :param dictionary: a dictionary to return values from
    :param key: key that specifies a value in the dictionary
    :return: generator for values that match the given key
    """
    def safe_ref(k, d):
        if d:
            try:
                return d[k]
            except KeyError:
                pass

    if not is_str(key):
        raise TypeError("key argument must of be of type 'str'")
    key_list = key.split('.')
    for k, le in last_element(dictionary, key_list):
        yield safe_ref(k, le)


def set_key_value(dictionary, key, value):
    """Set values all values in dictionary matching a dotstring key to a specified value.

    :param dictionary: a dictionary to set values in
    :param key: key that specifies an element in the dictionary
    :return: dictionary after changes have been made
    """
    def safe_assign(k, d):
        if d:
            try:
                d[k] = value
            except KeyError:
                pass

    if not is_str(key):
        raise TypeError("key argument must of be of type 'str'")
    key_list = key.split('.')
    for k, le in last_element(dictionary, key_list):
        safe_assign(k, le)
    return dictionary


def remove_key(dictionary, key):
    """Remove field specified by the docstring key

    :param dictionary: a dictionary to remove the value from
    :param key: key that specifies an element in the dictionary
    :return: dictionary after changes have been made
    """
    if not is_str(key):
        raise TypeError("key argument must of be of type 'str'")
    key_list = key.split('.')
    for k, le in last_element(dictionary, key_list):
        try:
            del le[k]
        except KeyError:
            pass
    return dictionary


def list_length(d, field):
    """Return the length of a list specified by field.

    If field represents a list in the document, then return its length.
    Otherwise return 0.

    :param d: a dictionary
    :param field: the dotstring field specifying a list
    """
    default_value = 0

    try:
        lst = next(key_value(d, field))
    except StopIterator:
        return default_value

    if isinstance(lst, list):
        return len(lst)
    else:
        return default_value



# -*- coding: utf-8 -*-
#
# python-json-patch - An implementation of the JSON Patch format
# https://github.com/stefankoegl/python-json-patch
#
# Copyright (c) 2011 Stefan Kögl <stefan@skoegl.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

""" Apply JSON-Patches (RFC 6902) """

from __future__ import unicode_literals

import collections
import copy
import functools
import inspect
import itertools
import json
import sys

try:
    from collections.abc import MutableMapping, MutableSequence
except ImportError:
    from collections import MutableMapping, MutableSequence

from jsonpointer import JsonPointer, JsonPointerException

# Will be parsed by setup.py to determine package metadata
__author__ = 'Stefan Kögl <stefan@skoegl.net>'
__version__ = '1.13'
__website__ = 'https://github.com/stefankoegl/python-json-patch'
__license__ = 'Modified BSD License'


# pylint: disable=E0611,W0404
if sys.version_info >= (3, 0):
    basestring = (bytes, str)  # pylint: disable=C0103,W0622


class JsonPatchException(Exception):
    """Base Json Patch exception"""


class InvalidJsonPatch(JsonPatchException):
    """ Raised if an invalid JSON Patch is created """


class JsonPatchConflict(JsonPatchException):
    """Raised if patch could not be applied due to conflict situation such as:
    - attempt to add object key then it already exists;
    - attempt to operate with nonexistence object key;
    - attempt to insert value to array at position beyond of it size;
    - etc.
    """


class JsonPatchTestFailed(JsonPatchException, AssertionError):
    """ A Test operation failed """


def multidict(ordered_pairs):
    """Convert duplicate keys values to lists."""
    # read all values into lists
    mdict = collections.defaultdict(list)
    for key, value in ordered_pairs:
        mdict[key].append(value)

    return dict(
        # unpack lists that have only 1 item
        (key, values[0] if len(values) == 1 else values)
        for key, values in mdict.items()
    )


def get_loadjson():
    """ adds the object_pairs_hook parameter to json.load when possible

    The "object_pairs_hook" parameter is used to handle duplicate keys when
    loading a JSON object. This parameter does not exist in Python 2.6. This
    methods returns an unmodified json.load for Python 2.6 and a partial
    function with object_pairs_hook set to multidict for Python versions that
    support the parameter. """

    if sys.version_info >= (3, 3):
        args = inspect.signature(json.load).parameters
    else:
        args = inspect.getargspec(json.load).args
    if 'object_pairs_hook' not in args:
        return json.load

    return functools.partial(json.load, object_pairs_hook=multidict)

json.load = get_loadjson()


def apply_patch(doc, patch, in_place=False, ignore_conflicts=False, verify=False):
    """Apply list of patches to specified json document.

    :param doc: Document object.
    :type doc: dict

    :param patch: JSON patch as list of dicts or raw JSON-encoded string.
    :type patch: list or str

    :param in_place: While :const:`True` patch will modify target document.
                     By default patch will be applied to document copy.
    :type in_place: bool

    :param ignore_conflicts: Ignore JsonConflicts errors
    :type ignore_conflicts: bool

    :param verify: works with `ignore_conflicts` = True, if errors and `verify` is
                  True (recommanded), make sure the resulting objects is the same
                  as the original one. `ignore_conflicts` and `verify` are used to 
                  run patches multiple times and get rif of errors when operations can't
                  be performed multiple times because the object has already been patched
                  This will force `in_place` to False in order the comparison to occur.
    :type verify: bool

    :return: Patched document object.
    :rtype: dict

    >>> doc = {'foo': 'bar'}
    >>> patch = [{'op': 'add', 'path': '/baz', 'value': 'qux'}]
    >>> other = apply_patch(doc, patch)
    >>> doc is not other
    True
    >>> other == {'foo': 'bar', 'baz': 'qux'}
    True
    >>> patch = [{'op': 'add', 'path': '/baz', 'value': 'qux'}]
    >>> apply_patch(doc, patch, in_place=True) == {'foo': 'bar', 'baz': 'qux'}
    True
    >>> doc == other
    True
    """

    if isinstance(patch, basestring):
        patch = JsonPatch.from_string(patch)
    else:
        patch = JsonPatch(patch)
    return patch.apply(doc, in_place, ignore_conflicts)

def reapply_patch(doc, patch):
    """Apply or (safely) re-apply patch to doc"""
    return apply_patch(doc,patch,ignore_conflicts=True, verify=True)

def make_patch(src, dst):
    """Generates patch by comparing of two document objects. Actually is
    a proxy to :meth:`JsonPatch.from_diff` method.

    :param src: Data source document object.
    :type src: dict

    :param dst: Data source document object.
    :type dst: dict

    >>> src = {'foo': 'bar', 'numbers': [1, 3, 4, 8]}
    >>> dst = {'baz': 'qux', 'numbers': [1, 4, 7]}
    >>> patch = make_patch(src, dst)
    >>> new = patch.apply(src)
    >>> new == dst
    True
    """
    return JsonPatch.from_diff(src, dst)


class JsonPatch(object):
    """A JSON Patch is a list of Patch Operations.

    >>> patch = JsonPatch([
    ...     {'op': 'add', 'path': '/foo', 'value': 'bar'},
    ...     {'op': 'add', 'path': '/baz', 'value': [1, 2, 3]},
    ...     {'op': 'remove', 'path': '/baz/1'},
    ...     {'op': 'test', 'path': '/baz', 'value': [1, 3]},
    ...     {'op': 'replace', 'path': '/baz/0', 'value': 42},
    ...     {'op': 'remove', 'path': '/baz/1'},
    ... ])
    >>> doc = {}
    >>> result = patch.apply(doc)
    >>> expected = {'foo': 'bar', 'baz': [42]}
    >>> result == expected
    True

    JsonPatch object is iterable, so you could easily access to each patch
    statement in loop:

    >>> lpatch = list(patch)
    >>> expected = {'op': 'add', 'path': '/foo', 'value': 'bar'}
    >>> lpatch[0] == expected
    True
    >>> lpatch == patch.patch
    True

    Also JsonPatch could be converted directly to :class:`bool` if it contains
    any operation statements:

    >>> bool(patch)
    True
    >>> bool(JsonPatch([]))
    False

    This behavior is very handy with :func:`make_patch` to write more readable
    code:

    >>> old = {'foo': 'bar', 'numbers': [1, 3, 4, 8]}
    >>> new = {'baz': 'qux', 'numbers': [1, 4, 7]}
    >>> patch = make_patch(old, new)
    >>> if patch:
    ...     # document have changed, do something useful
    ...     patch.apply(old)    #doctest: +ELLIPSIS
    {...}
    """
    def __init__(self, patch):
        self.patch = patch

        self.operations = {
            'remove': RemoveOperation,
            'add': AddOperation,
            'replace': ReplaceOperation,
            'move': MoveOperation,
            'test': TestOperation,
            'copy': CopyOperation,
        }

    def __str__(self):
        """str(self) -> self.to_string()"""
        return self.to_string()

    def __bool__(self):
        return bool(self.patch)

    __nonzero__ = __bool__

    def __iter__(self):
        return iter(self.patch)

    def __hash__(self):
        return hash(tuple(self._ops))

    def __eq__(self, other):
        if not isinstance(other, JsonPatch):
            return False
        return self._ops == other._ops

    def __ne__(self, other):
        return not(self == other)

    @classmethod
    def from_string(cls, patch_str):
        """Creates JsonPatch instance from string source.

        :param patch_str: JSON patch as raw string.
        :type patch_str: str

        :return: :class:`JsonPatch` instance.
        """
        patch = json.loads(patch_str)
        return cls(patch)

    @classmethod
    def from_diff(cls, src, dst):
        """Creates JsonPatch instance based on comparing of two document
        objects. Json patch would be created for `src` argument against `dst`
        one.

        :param src: Data source document object.
        :type src: dict

        :param dst: Data source document object.
        :type dst: dict

        :return: :class:`JsonPatch` instance.

        >>> src = {'foo': 'bar', 'numbers': [1, 3, 4, 8]}
        >>> dst = {'baz': 'qux', 'numbers': [1, 4, 7]}
        >>> patch = JsonPatch.from_diff(src, dst)
        >>> new = patch.apply(src)
        >>> new == dst
        True
        """
        def compare_values(path, value, other):
            if value == other:
                return
            if isinstance(value, MutableMapping) and \
                    isinstance(other, MutableMapping):
                for operation in compare_dicts(path, value, other):
                    yield operation
            elif isinstance(value, MutableSequence) and \
                    isinstance(other, MutableSequence):
                for operation in compare_lists(path, value, other):
                    yield operation
            else:
                ptr = JsonPointer.from_parts(path)
                yield {'op': 'replace', 'path': ptr.path, 'value': other}

        def compare_dicts(path, src, dst):
            for key in src:
                if key not in dst:
                    ptr = JsonPointer.from_parts(path + [key])
                    yield {'op': 'remove', 'path': ptr.path}
                    continue
                current = path + [key]
                for operation in compare_values(current, src[key], dst[key]):
                    yield operation
            for key in dst:
                if key not in src:
                    ptr = JsonPointer.from_parts(path + [key])
                    yield {'op': 'add',
                           'path': ptr.path,
                           'value': dst[key]}

        def compare_lists(path, src, dst):
            return _compare_lists(path, src, dst)

        return cls(list(compare_values([], src, dst)))

    def to_string(self):
        """Returns patch set as JSON string."""
        return json.dumps(self.patch)

    @property
    def _ops(self):
        return tuple(map(self._get_operation, self.patch))

    def apply(self, orig_obj, in_place=False, ignore_conflicts=False, verify=False):
        """Applies the patch to given object.

        :param obj: Document object.
        :type obj: dict

        :param in_place: Tweaks way how patch would be applied - directly to
                         specified `obj` or to his copy.
        :type in_place: bool

        :type ignore_conflicts: Ignore JsonConflicts errors

        :type verify: works with `ignore_conflicts` = True, if errors and `verify` is
                      True (recommanded), make sure the resulting objects is the same
                      as the original one. `ignore_conflicts` and `verify` are used to 
                      run patches multiple times and get rif of errors when operations can't
                      be performed multiple times because the object has already been patched

        :return: Modified `obj`.
        """

        if verify:
            in_place = False
        if not in_place:
            obj = copy.deepcopy(orig_obj)
        else:
            obj = orig_obj

        got_conflicts = False
        for operation in self._ops:
            try:
                obj = operation.apply(obj)
            except JsonPatchConflict as e:
                if ignore_conflicts:
                    got_conflicts = True
                else:
                    raise
        # it you're gonna ignore conflicts you'll have to make
        # sure the resulting document is the same as the passed-one
        # (patch run mutiple times)
        if got_conflicts and verify:
            assert obj == orig_obj, "Resulting object is different from original but got conflict errors, this is not good..."

        return obj

    def _get_operation(self, operation):
        if 'op' not in operation:
            raise InvalidJsonPatch("Operation does not contain 'op' member")

        op = operation['op']

        if not isinstance(op, basestring):
            raise InvalidJsonPatch("Operation must be a string")

        if op not in self.operations:
            raise InvalidJsonPatch("Unknown operation {0!r}".format(op))

        cls = self.operations[op]
        return cls(operation)


class PatchOperation(object):
    """A single operation inside a JSON Patch."""

    def __init__(self, operation):
        self.location = operation['path']
        self.pointer = JsonPointer(self.location)
        self.operation = operation

    def apply(self, obj):
        """Abstract method that applies patch operation to specified object."""
        raise NotImplementedError('should implement patch operation.')

    def __hash__(self):
        return hash(frozenset(self.operation.items()))

    def __eq__(self, other):
        if not isinstance(other, PatchOperation):
            return False
        return self.operation == other.operation

    def __ne__(self, other):
        return not(self == other)


class RemoveOperation(PatchOperation):
    """Removes an object property or an array element."""

    def apply(self, obj):
        subobj, part = self.pointer.to_last(obj)
        try:
            del subobj[part]
        except (KeyError, IndexError) as ex:
            msg = "can't remove non-existent object '{0}'".format(part)
            raise JsonPatchConflict(msg)

        return obj


class AddOperation(PatchOperation):
    """Adds an object property or an array element."""

    def apply(self, obj):
        try:
            value = self.operation["value"]
        except KeyError as ex:
            raise InvalidJsonPatch(
                "The operation does not contain a 'value' member")

        subobj, part = self.pointer.to_last(obj)

        if isinstance(subobj, MutableSequence):
            if part == '-':
                subobj.append(value)  # pylint: disable=E1103

            elif part > len(subobj) or part < 0:
                raise JsonPatchConflict("can't insert outside of list")

            else:
                subobj.insert(part, value)  # pylint: disable=E1103

        elif isinstance(subobj, MutableMapping):
            if part is None:
                obj = value  # we're replacing the root
            else:
                subobj[part] = value

        else:
            raise TypeError("invalid document type {0}".format(type(subobj)))

        return obj


class ReplaceOperation(PatchOperation):
    """Replaces an object property or an array element by new value."""

    def apply(self, obj):
        try:
            value = self.operation["value"]
        except KeyError as ex:
            raise InvalidJsonPatch(
                "The operation does not contain a 'value' member")

        subobj, part = self.pointer.to_last(obj)

        if part is None:
            return value

        if isinstance(subobj, MutableSequence):
            if part > len(subobj) or part < 0:
                raise JsonPatchConflict("can't replace outside of list")

        elif isinstance(subobj, MutableMapping):
            if not part in subobj:
                msg = "can't replace non-existent object '{0}'".format(part)
                raise JsonPatchConflict(msg)
        else:
            raise TypeError("invalid document type {0}".format(type(subobj)))

        subobj[part] = value
        return obj


class MoveOperation(PatchOperation):
    """Moves an object property or an array element to new location."""

    def apply(self, obj):
        try:
            from_ptr = JsonPointer(self.operation['from'])
        except KeyError as ex:
            raise InvalidJsonPatch(
                "The operation does not contain a 'from' member")

        subobj, part = from_ptr.to_last(obj)
        try:
            value = subobj[part]
        except (KeyError, IndexError) as ex:
            raise JsonPatchConflict(str(ex))

        # If source and target are equal, this is a no-op
        if self.pointer == from_ptr:
            return obj

        if isinstance(subobj, MutableMapping) and \
                self.pointer.contains(from_ptr):
            raise JsonPatchConflict('Cannot move values into its own children')

        obj = RemoveOperation({
            'op': 'remove',
            'path': self.operation['from']
        }).apply(obj)

        obj = AddOperation({
            'op': 'add',
            'path': self.location,
            'value': value
        }).apply(obj)

        return obj


class TestOperation(PatchOperation):
    """Test value by specified location."""

    def apply(self, obj):
        try:
            subobj, part = self.pointer.to_last(obj)
            if part is None:
                val = subobj
            else:
                val = self.pointer.walk(subobj, part)
        except JsonPointerException as ex:
            raise JsonPatchTestFailed(str(ex))

        try:
            value = self.operation['value']
        except KeyError as ex:
            raise InvalidJsonPatch(
                "The operation does not contain a 'value' member")

        if val != value:
            msg = '{0} ({1}) is not equal to tested value {2} ({3})'
            raise JsonPatchTestFailed(msg.format(val, type(val),
                                                 value, type(value)))

        return obj


class CopyOperation(PatchOperation):
    """ Copies an object property or an array element to a new location """

    def apply(self, obj):
        try:
            from_ptr = JsonPointer(self.operation['from'])
        except KeyError as ex:
            raise InvalidJsonPatch(
                "The operation does not contain a 'from' member")

        subobj, part = from_ptr.to_last(obj)
        try:
            value = copy.deepcopy(subobj[part])
        except (KeyError, IndexError) as ex:
            raise JsonPatchConflict(str(ex))

        obj = AddOperation({
            'op': 'add',
            'path': self.location,
            'value': value
        }).apply(obj)

        return obj


def _compare_lists(path, src, dst):
    """Compares two lists objects and return JSON patch about."""
    return _optimize(_compare(path, src, dst, *_split_by_common_seq(src, dst)))


def _longest_common_subseq(src, dst):
    """Returns pair of ranges of longest common subsequence for the `src`
    and `dst` lists.

    >>> src = [1, 2, 3, 4]
    >>> dst = [0, 1, 2, 3, 5]
    >>> # The longest common subsequence for these lists is [1, 2, 3]
    ... # which is located at (0, 3) index range for src list and (1, 4) for
    ... # dst one. Tuple of these ranges we should get back.
    ... assert ((0, 3), (1, 4)) == _longest_common_subseq(src, dst)
    """
    lsrc, ldst = len(src), len(dst)
    drange = list(range(ldst))
    matrix = [[0] * ldst for _ in range(lsrc)]
    z = 0  # length of the longest subsequence
    range_src, range_dst = None, None
    for i, j in itertools.product(range(lsrc), drange):
        if src[i] == dst[j]:
            if i == 0 or j == 0:
                matrix[i][j] = 1
            else:
                matrix[i][j] = matrix[i-1][j-1] + 1
            if matrix[i][j] > z:
                z = matrix[i][j]
            if matrix[i][j] == z:
                range_src = (i-z+1, i+1)
                range_dst = (j-z+1, j+1)
        else:
            matrix[i][j] = 0
    return range_src, range_dst


def _split_by_common_seq(src, dst, bx=(0, -1), by=(0, -1)):
    """Recursively splits the `dst` list onto two parts: left and right.
    The left part contains differences on left from common subsequence,
    same as the right part by for other side.

    To easily understand the process let's take two lists: [0, 1, 2, 3] as
    `src` and [1, 2, 4, 5] for `dst`. If we've tried to generate the binary tree
    where nodes are common subsequence for both lists, leaves on the left
    side are subsequence for `src` list and leaves on the right one for `dst`,
    our tree would looks like::

        [1, 2]
       /     \
    [0]       []
             /  \
          [3]   [4, 5]

    This function generate the similar structure as flat tree, but without
    nodes with common subsequences - since we're don't need them - only with
    left and right leaves::

        []
       / \
    [0]  []
        / \
     [3]  [4, 5]

    The `bx` is the absolute range for currently processed subsequence of
    `src` list.  The `by` means the same, but for the `dst` list.
    """
    # Prevent useless comparisons in future
    bx = bx if bx[0] != bx[1] else None
    by = by if by[0] != by[1] else None

    if not src:
        return [None, by]
    elif not dst:
        return [bx, None]

    # note that these ranges are relative for processed sublists
    x, y = _longest_common_subseq(src, dst)

    if x is None or y is None:  # no more any common subsequence
        return [bx, by]

    return [_split_by_common_seq(src[:x[0]], dst[:y[0]],
                                 (bx[0], bx[0] + x[0]),
                                 (by[0], by[0] + y[0])),
            _split_by_common_seq(src[x[1]:], dst[y[1]:],
                                 (bx[0] + x[1], bx[0] + len(src)),
                                 (by[0] + y[1], by[0] + len(dst)))]


def _compare(path, src, dst, left, right):
    """Same as :func:`_compare_with_shift` but strips emitted `shift` value."""
    for op, _ in _compare_with_shift(path, src, dst, left, right, 0):
        yield op


def _compare_with_shift(path, src, dst, left, right, shift):
    """Recursively compares differences from `left` and `right` sides
    from common subsequences.

    The `shift` parameter is used to store index shift which caused
    by ``add`` and ``remove`` operations.

    Yields JSON patch operations and list index shift.
    """
    if isinstance(left, MutableSequence):
        for item, shift in _compare_with_shift(path, src, dst, *left,
                                               shift=shift):
            yield item, shift
    elif left is not None:
        for item, shift in _compare_left(path, src, left, shift):
            yield item, shift

    if isinstance(right, MutableSequence):
        for item, shift in _compare_with_shift(path, src, dst, *right,
                                               shift=shift):
            yield item, shift
    elif right is not None:
        for item, shift in _compare_right(path, dst, right, shift):
            yield item, shift


def _compare_left(path, src, left, shift):
    """Yields JSON patch ``remove`` operations for elements that are only
    exists in the `src` list."""
    start, end = left
    if end == -1:
        end = len(src)
    # we need to `remove` elements from list tail to not deal with index shift
    for idx in reversed(range(start + shift, end + shift)):
        ptr = JsonPointer.from_parts(path + [str(idx)])
        yield (
            {'op': 'remove',
             # yes, there should be any value field, but we'll use it
             # to apply `move` optimization a bit later and will remove
             # it in _optimize function.
             'value': src[idx - shift],
             'path': ptr.path,
            },
            shift - 1
        )
        shift -= 1


def _compare_right(path, dst, right, shift):
    """Yields JSON patch ``add`` operations for elements that are only
    exists in the `dst` list"""
    start, end = right
    if end == -1:
        end = len(dst)
    for idx in range(start, end):
        ptr = JsonPointer.from_parts(path + [str(idx)])
        yield (
            {'op': 'add', 'path': ptr.path, 'value': dst[idx]},
            shift + 1
        )
        shift += 1


def _optimize(operations):
    """Optimizes operations which was produced by lists comparison.

    Actually it does two kinds of optimizations:

    1. Seeks pair of ``remove`` and ``add`` operations against the same path
       and replaces them with ``replace`` operation.
    2. Seeks pair of ``remove`` and ``add`` operations for the same value
       and replaces them with ``move`` operation.
    """
    result = []
    ops_by_path = {}
    ops_by_value = {}
    add_remove = set(['add', 'remove'])
    for item in operations:
        # could we apply "move" optimization for dict values?
        hashable_value = not isinstance(item['value'],
                                        (MutableMapping, MutableSequence))
        if item['path'] in ops_by_path:
            _optimize_using_replace(ops_by_path[item['path']], item)
            continue
        if hashable_value and item['value'] in ops_by_value:
            prev_item = ops_by_value[item['value']]
            # ensure that we processing pair of add-remove ops
            if set([item['op'], prev_item['op']]) == add_remove:
                _optimize_using_move(prev_item, item)
                ops_by_value.pop(item['value'])
                continue
        result.append(item)
        ops_by_path[item['path']] = item
        if hashable_value:
            ops_by_value[item['value']] = item

    # cleanup
    ops_by_path.clear()
    ops_by_value.clear()
    for item in result:
        if item['op'] == 'remove':
            item.pop('value')  # strip our hack
        yield item


def _optimize_using_replace(prev, cur):
    """Optimises by replacing ``add``/``remove`` with ``replace`` on same path

    For nested strucures, tries to recurse replacement, see #36 """
    prev['op'] = 'replace'
    if cur['op'] == 'add':
        # make recursive patch
        patch = make_patch(prev['value'], cur['value'])
        if len(patch.patch) == 1 and patch.patch[0]['op'] != 'remove':
            prev['path'] = prev['path'] + patch.patch[0]['path']
            prev['value'] = patch.patch[0]['value']
        else:
            prev['value'] = cur['value']


def _optimize_using_move(prev_item, item):
    """Optimises JSON patch by using ``move`` operation instead of
    ``remove` and ``add`` against the different paths but for the same value."""
    prev_item['op'] = 'move'
    move_from, move_to = [
        (item['path'], prev_item['path']),
        (prev_item['path'], item['path']),
    ][item['op'] == 'add']
    if item['op'] == 'add':  # first was remove then add
        prev_item['from'] = move_from
        prev_item['path'] = move_to
    else:  # first was add then remove
        head, move_from = move_from.rsplit('/', 1)
        # since add operation was first it incremented
        # overall index shift value. we have to fix this
        move_from = int(move_from) - 1
        prev_item['from'] = head + '/%d' % move_from
        prev_item['path'] = move_to



from multiprocessing import Pool
from glob import glob
from os.path import abspath, isdir, join, exists
from os import cpu_count
from traceback import format_exc

from biothings.utils.backend import DocBackendOptions
from biothings.utils.common import iter_n


# How many cpus on this machine?
DEFAULT_THREADS = cpu_count()

# simple aggregation functions
def agg_by_sum(prev, curr):
    return prev + curr

def agg_by_append(prev, curr):
    if isinstance(curr, list):
        return prev + curr
    return prev + [curr]

# avoid the global variable and the callback function this way
class ParallelResult(object):
    def __init__(self, agg_function, agg_function_init):
        self.res = agg_function_init
        self.agg_function = agg_function

    def aggregate(self, curr):
        self.res = self.agg_function(self.res, curr)

# Handles errors in async apply
class ErrorHandler(object):
    def __init__(self, errpath, chunk_num):
        if errpath:
            self.error_file_path = errpath + '_{}'.format(chunk_num)
        else:
            self.error_file_path = None

    def handle(self, exception):
        if self.error_file_path:
            f = open(self.error_file_path, 'w')
            f.write(format_exc())
            f.write('\n{}\n'.format(str(exception)))
            f.close()

def run_parallel_on_iterable(fun, iterable, backend_options=None, agg_function=agg_by_append, agg_function_init=[],
                             chunk_size=1000000, num_workers=DEFAULT_THREADS, outpath=None,
                             mget_chunk_size=10000, ignore_None=True, error_path=None, **query_kwargs):

    ''' This function will run a user function on all documents in a backend database in parallel using
        multiprocessing.Pool.  The overview of the process looks like this:

        Chunk (into chunks of size "chunk_size") items in iterable, and run the following
        script on each chunk using a multiprocessing.Pool object with "num_workers" processes:
            For each document in list of ids in this chunk (documents retrived in chunks of "mget_chunk_size"):
                Run function "fun" with parameters
                (doc, chunk_num, f <file handle only passed if "outpath" is not None>),
                and aggregate the result with the current results using function "agg_function".

        :param fun:
                The function to run on all documents.  If outpath is NOT specified, fun
                must accept two parameters: (doc, chunk_num), where doc is the backend document,
                and chunk_num is essentially a unique process id.
                If outpath IS specified, an additional open file handle (correctly tagged with
                the current chunk's chunk_num) will also be passed to fun, and thus it
                must accept three parameters: (doc, chunk_num, f)
        :param iterable:
                Iterable of ids.
        :param backend_options:
                An instance of biothings.utils.backend.DocBackendOptions.  This contains
                the options necessary to instantiate the correct backend class (ES, mongo, etc).
        :param agg_function:
                This function aggregates the return value of each run of function fun.  It should take
                2 parameters: (prev, curr), where prev is the previous aggregated result, and curr
                is the output of the current function run.  It should return some value that represents
                the aggregation of the previous aggregated results with the output of the current
                function.
        :param agg_function_init:
                Initialization value for the aggregated result.
        :param chunk_size:
                Length of the ids list sent to each chunk.
        :param num_workers:
                Number of processes that consume chunks in parallel.
                https://docs.python.org/2/library/multiprocessing.html#multiprocessing.pool.multiprocessing.Pool
        :param outpath:
                Base path for output files.  Because function fun can be run many times in parallel, each
                chunk is sequentially numbered, and the output file name for any chunk is outpath_{chunk_num},
                e.g., if outpath is out, all output files will be of the form: /path/to/cwd/out_1,
                /path/to/cwd/out_2, etc.
        :param error_path:
                Base path for error files.  If included, exceptions inside each chunk thread will be printed to
                these files.
        :param mget_chunk_size:
                The size of each mget chunk inside each chunk thread.  In each thread, the ids list
                is consumed by passing chunks to a mget_by_ids function.  This parameter controls the size of
                each mget.
        :param ignore_None:
                If set, then falsy values will not be aggregated (0, [], None, etc) in the aggregation step.
                Default True.

        All other parameters are fed to the backend query.
      '''

    ret = ParallelResult(agg_function, agg_function_init)

    # assert backend_options is correct
    if not backend_options or not isinstance(backend_options, DocBackendOptions):
        raise Exception("backend_options must be a biothings.databuild.parallel2.DocBackendOptions class")

    # build backend from options
    backend = backend_options.cls.create_from_options(backend_options)

    # normalize path for out files
    if outpath:
        outpath = abspath(outpath)

    if error_path:
        error_path = abspath(error_path)

    with Pool(processes=num_workers) as p:
        for (chunk_num, chunk) in enumerate(iter_n(iterable, chunk_size)):
            # apply function to chunk
            p.apply_async(
                _run_one_chunk,
                args=(chunk_num, chunk, fun, backend_options, agg_function, agg_function_init,
                      outpath, mget_chunk_size, ignore_None),
                callback=ret.aggregate,
                error_callback=ErrorHandler(error_path, chunk_num).handle
            )
        # close pool and wait for completion of all workers
        p.close()
        p.join()
    return ret.res


def run_parallel_on_ids_file(fun, ids_file, backend_options=None, agg_function=agg_by_append, agg_function_init=[],
                             chunk_size=1000000, num_workers=DEFAULT_THREADS, outpath=None,
                             mget_chunk_size=10000, ignore_None=True, error_path=None, **query_kwargs):
    ''' Implementation of run_parallel_on_iterable, where iterable comes from the lines of a file.

        All parameters are fed to run_on_ids_iterable, except:

        :param ids_file:    Path to file with ids, one per line.
'''

    def _file_iterator(fh):
        for line in fh:
            yield line.strip('\n')

    ids_file = abspath(ids_file)
    with open(ids_file, 'r') as ids_handle:
        return run_parallel_on_iterable(
            fun=fun, iterable=_file_iterator(ids_handle),
            backend_options=backend_options, agg_function=agg_function,
            agg_function_init=agg_function_init, chunk_size=chunk_size,
            num_workers=num_workers, outpath=outpath, mget_chunk_size=mget_chunk_size,
            ignore_None=ignore_None, error_path=error_path, **query_kwargs
        )

# TODO: allow mget args to be passed
def run_parallel_on_query(fun, backend_options=None, query=None, agg_function=agg_by_append,
                          agg_function_init=[], chunk_size=1000000, num_workers=DEFAULT_THREADS,
                          outpath=None, mget_chunk_size=10000, ignore_None=True, error_path=None,
                          full_doc=False, **query_kwargs):
    ''' Implementation of run_parallel_on_ids_iterable, where the ids iterable comes from the result of a query
        on the specified backend.

        All parameters are fed to run_parallel_on_ids_iterable, except:

        :param query:       ids come from results of this query run on backend, default: "match_all"
        :param full_doc:    If True, a list of documents is passed to each subprocess, rather than ids that are
                            looked up later.  Should be faster?
                            Unknown how this works with very large query sets...
    '''
    def _query_iterator(q):
        for doc in q:
            if full_doc:
                yield doc
            else:
                yield doc['_id']

    # assert backend_options is correct
    if not backend_options or not isinstance(backend_options, DocBackendOptions):
        raise Exception("backend_options must be a biothings.databuild.parallel2.DocBackendOptions class")

    # build backend from options
    backend = backend_options.cls.create_from_options(backend_options)

    return run_parallel_on_iterable(
        fun=fun,
        iterable=_query_iterator(backend.query(query, _source=full_doc, only_source=False, **query_kwargs)),
        backend_options=backend_options, agg_function=agg_function,
        agg_function_init=agg_function_init, chunk_size=chunk_size, outpath=outpath,
        num_workers=num_workers, mget_chunk_size=mget_chunk_size, ignore_None=ignore_None,
        error_path=error_path, **query_kwargs
    )

def run_parallel_on_ids_dir(fun, ids_dir, backend_options=None, agg_function=agg_by_append,
                            agg_function_init=[], outpath=None,
                            num_workers=DEFAULT_THREADS, mget_chunk_size=10000,
                            ignore_None=True, error_path=None, **query_kwargs):
    ''' This function will run function fun on chunks defined by the files in ids_dir.

        All parameters are fed to run_parallel_on_iterable, except:

        :params ids_dir:    Directory containing only files with ids, one per line.  The number of files
                            defines the number of chunks.

    '''
    ids_dir = abspath(ids_dir)
    assert isdir(ids_dir)
    return run_parallel_on_iterable(
        fun=fun, iterable=glob(join(ids_dir, '*')),
        backend_options=backend_options, agg_function=agg_function,
        agg_function_init=agg_function_init, chunk_size=1, outpath=outpath,
        num_workers=num_workers, mget_chunk_size=mget_chunk_size,
        ignore_None=ignore_None, error_path=error_path, **query_kwargs
    )


def _run_one_chunk(chunk_num, chunk, fun, backend_options, agg_function,
                   agg_function_init, outpath, mget_chunk_size, ignore_None):
    # iterator if chunk is a file path
    def _file_path_iterator(fp):
        with open(fp, 'r') as fh:
            for line in fh:
                yield line.strip('\n')

    # recreate backend object
    backend = backend_options.cls.create_from_options(backend_options)

    # check to see chunk is a list of ids, a list of objects, or a tuple with a path and make
    # the correct iterator for the type
    if isinstance(chunk, list) or isinstance(chunk, tuple):
        if isinstance(chunk[0], str) and exists(chunk[0]):
            iterator = backend.mget_from_ids(_file_path_iterator(chunk[0]), step=mget_chunk_size, only_source=False)
        elif isinstance(chunk[0], str):
            iterator = backend.mget_from_ids(chunk, step=mget_chunk_size, only_source=False)
        elif isinstance(chunk[0], dict):
            iterator = iter(chunk)
        else:
            raise Exception("lists and tuples can contain either strings (containing ids or a file path), or dicts (containing full documents)")
    else:
        raise Exception("chunk can only be a list/tuple of ids, a list/tuple of objects, or a list/tuple with a file path to a list of ids.")

    # initialize return for this chunk
    ret = ParallelResult(agg_function, agg_function_init)

    # make file handle for this chunk
    if outpath:
        _file = open(outpath + '_{}'.format(chunk_num), 'w')

    # iterate through this chunk and run function on every doc
    for doc in iterator:
        # Actually call the function
        if outpath:
            r = fun(doc, chunk_num, _file)
        else:
            r = fun(doc, chunk_num)

        # aggregate the results
        if r or not ignore_None:
            ret.aggregate(r)

    # close handle
    if outpath and _file:
        _file.close()

    return ret.res



import os
import sqlite3
import json

from biothings.utils.hub_db import IDatabase
from biothings.utils.dotfield import parse_dot_fields
from biothings.utils.dataload import update_dict_recur
from biothings.utils.common import json_serial

def get_hub_db_conn():
    return Database()

def get_src_dump():
    db = Database()
    return db[getattr(db.CONFIG, "DATA_SRC_DUMP_COLLECTION", "src_dump")]

def get_src_master():
    db = Database()
    return db[db.CONFIG.DATA_SRC_MASTER_COLLECTION]

def get_src_build():
    db = Database()
    return db[db.CONFIG.DATA_SRC_BUILD_COLLECTION]

def get_src_build_config():
    db = Database()
    return db[db.CONFIG.DATA_SRC_BUILD_CONFIG_COLLECTION]

def get_data_plugin():
    db = Database()
    return db[db.CONFIG.DATA_PLUGIN_COLLECTION]

def get_api():
    db = Database()
    return db[db.CONFIG.API_COLLECTION]

def get_cmd():
    db = Database()
    return db[db.CONFIG.CMD_COLLECTION]

def get_event():
    db = Database()
    return db[getattr(db.CONFIG, "EVENT_COLLECTION", "event")]

def get_hub_config():
    db = Database()
    return db[getattr(db.CONFIG, "HUB_CONFIG_COLLECTION", "hub_config")]

def get_last_command():
    try:
        db = Database()
        res = db.get_conn().execute("SELECT MAX(_id) FROM cmd").fetchall()
        assert res[0][0], "No command ID found, bootstrap ?"
        return {"_id": res[0][0]}
    except Exception:
        return {"_id": 1}

def get_source_fullname(col_name):
    """
    Assuming col_name is a collection created from an upload process,
    find the main source & sub_source associated.
    """
    src_dump = get_src_dump()
    info = None
    for doc in src_dump.find():
        if col_name in doc.get("upload", {}).get("jobs", {}).keys():
            info = doc
    if info:
        name = info["_id"]
        if name != col_name:
            # col_name was a sub-source name
            return "%s.%s" % (name, col_name)
        else:
            return name

class Database(IDatabase):

    def __init__(self):
        super(Database, self).__init__()
        self.name = getattr(self.CONFIG, "DATA_HUB_DB_DATABASE", "hubdb")
        if not os.path.exists(self.CONFIG.HUB_DB_BACKEND["sqlite_db_folder"]):
            os.makedirs(self.CONFIG.HUB_DB_BACKEND["sqlite_db_folder"])
        self.dbfile = os.path.join(self.CONFIG.HUB_DB_BACKEND["sqlite_db_folder"], self.name)
        self.cols = {}

    @property
    def address(self):
        return self.dbfile

    def get_conn(self):
        return sqlite3.connect(self.dbfile)

    def collection_names(self):
        tables = self.get_conn().execute("SELECT name FROM sqlite_master WHERE type='table'").fetchall()
        return [name[0] for name in tables]

    def create_collection(self, colname):
        return self[colname]

    def create_if_needed(self, table):
        existings = [tname[0] for tname in self.get_conn().execute(
            "SELECT name FROM sqlite_master WHERE type='table' and " + "name = ?",
            (table,)
        ).fetchall()]
        if table not in existings:
            # TODO: injection...
            self.get_conn().execute("CREATE TABLE %s (_id TEXT PRIMARY KEY, document TEXT)" % table).fetchone()

    def __getitem__(self, colname):
        if colname not in self.cols:
            self.create_if_needed(colname)
            self.cols[colname] = Collection(colname, self)
        return self.cols[colname]


class Collection(object):

    def __init__(self, colname, db):
        self.colname = colname
        self.db = db

    def get_conn(self):
        return sqlite3.connect(self.db.dbfile)

    @property
    def name(self):
        return self.colname

    @property
    def database(self):
        return self.db

    def find_one(self, *args, **kwargs):
        if args and len(args) == 1 and isinstance(args[0], dict):
            if len(args[0]) == 1 and "_id" in args[0]:
                strdoc = self.get_conn().execute("SELECT document FROM %s WHERE _id = ?" % self.colname, (args[0]["_id"],)).fetchone()
                if strdoc:
                    return json.loads(strdoc[0])
                else:
                    return None
            else:
                return self.find(*args, find_one=True)
        elif args or kwargs:
            raise NotImplementedError("find(): %s %s" % (repr(args), repr(kwargs)))
        else:
            return self.find(find_one=True)

    def find(self, *args, **kwargs):
        results = []
        if args and len(args) == 1 and isinstance(args[0], dict) and len(args[0]) > 0:
            # it's key/value search, let's iterate
            for doc in self.get_conn().execute("SELECT document FROM %s" % self.colname).fetchall():
                found = False
                doc = json.loads(doc[0])
                for k, v in args[0].items():
                    if k in doc:
                        if doc[k] == v:
                            found = True
                    else:
                        found = False
                        break
                if found:
                    if "find_one" in kwargs:
                        return doc
                    else:
                        results.append(doc)
            return results
        elif not args or len(args) == 1 and len(args[0]) == 0:
            # nothing or empty dict
            return [json.loads(doc[0]) for doc in
                    self.get_conn().execute("SELECT document FROM %s" % self.colname).fetchall()]
        else:
            raise NotImplementedError("find: args=%s kwargs=%s" % (repr(args), repr(kwargs)))

    def insert_one(self, doc):
        assert "_id" in doc
        with self.get_conn() as conn:
            conn.execute(
                "INSERT INTO %s (_id,document) VALUES (?,?)" % self.colname,
                (doc["_id"], json.dumps(doc, default=json_serial))
            ).fetchone()
            conn.commit()

    def update_one(self, query, what, upsert=False):
        assert len(what) == 1 and ("$set" in what or "$unset" in what or "$push" in what), "$set/$unset/$push operators not found"
        doc = self.find_one(query)
        if doc:
            if "$set" in what:
                # parse_dot_fields uses json.dumps internally, we can to make
                # sure everything is serializable first
                what = json.loads(json.dumps(what, default=json_serial))
                what = parse_dot_fields(what["$set"])
                doc = update_dict_recur(doc, what)
            elif "$unset" in what:
                for keytounset in what["$unset"].keys():
                    doc.pop(keytounset, None)
            elif "$push" in what:
                for listkey, elem in what["$push"].items():
                    assert "." not in listkey, "$push not supported for nested keys: %s" % listkey
                    doc.setdefault(listkey, []).append(elem)
            self.save(doc)
        elif upsert:
            assert "$set" in what
            query.update(what["$set"])
            self.save(query)

    def update(self, query, what):
        docs = self.find(query)
        for doc in docs:
            self.update_one({"_id": doc["_id"]}, what)

    def save(self, doc):
        if self.find_one({"_id": doc["_id"]}):
            with self.get_conn() as conn:
                conn.execute(
                    "UPDATE %s SET document = ? WHERE _id = ?" % self.colname,
                    (json.dumps(doc, default=json_serial), doc["_id"])
                )
                conn.commit()
        else:
            self.insert_one(doc)

    def replace_one(self, query, doc, upsert=False):
        assert "_id" in query
        orig = self.find_one(query)
        if orig:
            orig["_id"] = query["_id"]
            with self.get_conn() as conn:
                conn.execute(
                    "UPDATE %s SET document = ? WHERE _id = ?" % self.colname,
                    (json.dumps(doc, default=json_serial), orig["_id"])
                )
                conn.commit()
        elif upsert:
            doc["_id"] = query["_id"]
            self.save(doc)

    def remove(self, query):
        docs = self.find(query)
        with self.get_conn() as conn:
            for doc in docs:
                conn.execute("DELETE FROM %s WHERE _id = ?" % self.colname, (doc["_id"],)).fetchone()
            conn.commit()

    def count(self):
        return self.get_conn().execute("SELECT count(_id) FROM %s" % self.colname).fetchone()[0]

    def drop(self):
        self.get_conn().execute("DROP TABLE %s" % self.colname).fetchall()

    def __getitem__(self, _id):
        return self.find_one({"_id": _id})

    def __getstate__(self):
        self.__dict__.pop("db", None)
        return self.__dict__



import json

import bson

from biothings.utils.common import is_str


def generate_json_schema(dmap):

    scalarmap = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        bson.int64.Int64: "number",
        None: "null",
    }

    def merge_type(typ1, typ2):
        if isinstance(typ1, list):
            if isinstance(typ2, list):
                typ1.extend(typ2)
            else:
                typ1.append(typ2)
        elif isinstance(typ2, list):
            typ1 = [typ1] + typ1
        else:
            typ1 = [typ1, typ2]

        return list(set(typ1))

    schema = {}

    if isinstance(dmap, dict):
        for k in dmap:
            if is_str(k):
                esch = generate_json_schema(dmap[k])
                if schema:
                    if schema["type"] == "object":
                        # we just complete 'properties', key already defined previously
                        pass
                    elif schema["type"] == "array":
                        if not schema.get("properties"):
                            schema["properties"] = {}
                            schema["type"] = merge_type(schema["type"], "object")
                    elif isinstance(schema["type"], list):
                        assert set(schema["type"]) == {"object", "array"}
                    else:
                        raise Exception("Previous schema type not expected: %s" % schema["type"])

                else:
                    schema = {"type": "object", "properties": {}}
                schema["properties"][k] = esch
            # elif type(k) == type:    # TODO: remove this line
            elif isinstance(k, type):
                if k == list:
                    if schema:
                        # already defined for this key, mixed types
                        schema.update({"items": {}})
                        schema["type"] = merge_type(schema["type"], "array")
                    else:
                        schema = {"type": "array", "items": {}}
                    esch = generate_json_schema(dmap[k])
                    schema["items"] = generate_json_schema(dmap[k])
                else:
                    if schema:
                        schema["type"] = merge_type(schema["type"], scalarmap[k])
                    else:
                        schema = {"type": scalarmap[k]}
            elif k is None:
                schema = {"type": None}
            else:
                raise Exception("no not here, k: %s" % k)
    else:
        pass

    return schema


def test():
    # TODO: Move these test to tests folder, or maybe already moved over? Chunlei
    from biothings.utils.inspect import typify_inspect_doc, inspect_docs

    # can't use assert directly, as we can't ensure the order of types (for instance)
    import biothings.utils.jsondiff
    biothings.utils.jsondiff.UNORDERED_LIST = True
    jsondiff = biothings.utils.jsondiff.make

    # object
    td1 = {"i": {"a": 456}}
    s1 = {
        'properties': {
            'i': {
                'properties': {'a': {'type': 'integer'}},
                'type': 'object'
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td1], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s1) == [], "%s  !=\n%s" % (gs, s1)

    td5 = {"i": [1, 2, 3]}
    s5 = {
        'properties': {
            'i': {
                'items': {'type': 'integer'},
                'type': 'array'
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td5], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s5) == [], "%s  !=\n%s" % (gs, s5)

    # array of object
    td2 = {"i": [{"a": 123}]}
    s2 = {
        'properties': {
            'i': {
                'items': {
                    'properties': {'a': {'type': 'integer'}},
                    'type': 'object'
                },
                'type': 'array'
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td2], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s2) == [], "%s  !=\n%s" % (gs, s2)

    # object in object
    td3 = {"i": {"a": {"b": 123}}}
    s3 = {
        'properties': {
            'i': {
                'properties': {
                    'a': {
                        'properties': {'b': {'type': 'integer'}},
                        'type': 'object'
                    }
                },
                'type': 'object'
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td3], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s3) == [], "%s  !=\n%s" % (gs, s3)

    # mixed str/float in array
    td6 = {"i": [1, 2, "a"]}
    s6 = {
        'properties': {
            'i': {
                'items': {'type': ['integer', 'string']},
                'type': 'array'
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td6], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s6) == [], "%s  !=\n%s" % (gs, s6)

    # mixed array/object
    td1 = {"i": {"a": 456}}
    td2 = {"i": [{"a": 123}]}
    s12 = {
        'properties': {
            'i': {
                'items': {
                    'properties': {'a': {'type': 'integer'}},
                    'type': 'object'
                },
                'properties': {'a': {'type': 'integer'}},
                'type': ['array', 'object']
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td1, td2], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s12) == [], "%s  !=\n%s" % (gs, s12)

    # list of integer (list of things which are not objects)
    td4 = {'a': [5, 5, 3]}
    s4 = {
        'properties': {
            'a': {
                'items': {'type': 'integer'},
                'type': 'array'
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td4], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s4) == [], "%s  !=\n%s" % (gs, s4)

    td7 = {"i": {"a": 1, "b": 2}}
    s7 = {
        'type': 'object',
        'properties': {
            'i': {
                'type': 'object',
                'properties': {
                    'a': {'type': 'integer'},
                    'b': {'type': 'integer'}
                }
            }
        }
    }
    m = inspect_docs([td7], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s7) == [], "%s  !=\n%s" % (gs, s7)

    # int or list of int (not a list of dict, testing scalar there)
    td81 = {"i": 1}
    td82 = {"i": [2, 3]}
    s812 = {
        'properties': {
            'i': {
                'items': {'type': 'integer'},
                'type': ['array', 'integer']
            }
        },
        'type': 'object'
    }
    m = inspect_docs([td81, td82], mode="type")["type"]
    gs = generate_json_schema(m)
    assert jsondiff(gs, s812) == [], "%s  !=\n%s" % (gs, s812)

    # run from app folder, biothings as symlink

    # small real-life collection
    cgi_schema = json.load(open("biothings/tests/cgi_schema.json"))
    cgi_map = typify_inspect_doc(json.load(open("biothings/tests/cgi_map.json")))
    schema = generate_json_schema(cgi_map)
    assert jsondiff(cgi_schema, schema) == []

    clinvar_schema = json.load(open("biothings/tests/clinvar_schema.json"))
    clinvar_map = typify_inspect_doc(json.load(open("biothings/tests/clinvar_map.json")))
    schema = generate_json_schema(clinvar_map)
    assert jsondiff(clinvar_schema, schema) == []

    mygene_schema = json.load(open("biothings/tests/mygene_schema.json"))
    mygene_map = typify_inspect_doc(json.load(open("biothings/tests/mygene_map.json")))
    schema = generate_json_schema(mygene_map)
    assert jsondiff(mygene_schema, schema) == []

    print("All test OK")


if __name__ == "__main__":
    test()



import json
import logging
from collections import defaultdict

import elasticsearch
import elasticsearch_dsl
from elasticsearch_dsl.connections import Connections
from tornado.ioloop import IOLoop

from biothings.utils.web.es import get_es_versions
from biothings.utils.web.es_transport import (BiothingsAsyncTransport,
                                              BiothingsTransport)
from biothings.utils.web.run import run_once
from biothings.utils.web.userquery import ESUserQuery
from biothings.web.settings import BiothingWebSettings

should_log_es_py_ver = run_once()  # evaluate to true once
should_log_es_host_ver = run_once()  # evaluate to true once per host

class BiothingESWebSettings(BiothingWebSettings):
    '''
    `BiothingWebSettings`_ subclass with functions specific to an elasticsearch backend.
    '''

    def __init__(self, config=None, parent=None, **kwargs):
        '''
        The ``config`` init parameter specifies a module that configures
        this biothing.  For more information see `config module`_ documentation.
        '''
        super(BiothingESWebSettings, self).__init__(config, parent, **kwargs)

        # elasticsearch connections
        self._connections = Connections()

        connection_settings = {
            "hosts": self.ES_HOST,
            "timeout": self.ES_CLIENT_TIMEOUT
        }
        connection_settings.update(transport_class=BiothingsTransport)
        self._connections.create_connection(alias='sync', **connection_settings)
        connection_settings.update(transport_class=BiothingsAsyncTransport)
        if self.ES_SNIFF:
            connection_settings.update(sniffer_timeout=60)
            connection_settings.update(sniff_on_start=True)
            connection_settings.update(sniff_on_connection_fail=True)
        self._connections.create_connection(alias='async', **connection_settings)

        # cached index mappings
        self.source_metadata = defaultdict(dict)
        self.source_properties = defaultdict(dict)

        # populate field notes if exist
        try:
            inf = open(self.AVAILABLE_FIELDS_NOTES_PATH, 'r')
            self._fields_notes = json.load(inf)
            inf.close()
        except Exception:
            self._fields_notes = {}

        # user query data
        self.userquery = ESUserQuery(self.USERQUERY_DIR)

        # query pipelines
        self.query_builder = self.load_class(self.ES_QUERY_BUILDER)(self)
        self.query_backend = self.load_class(self.ES_QUERY_BACKEND)(self)
        self.result_transform = self.load_class(self.ES_RESULT_TRANSFORM)(self)

        # initialize payload for standalone tracking batch
        self.tracking_payload = []

        self.ES_INDICES = dict(self.ES_INDICES)  # TODO
        self.ES_INDICES[self.ES_DOC_TYPE] = self.ES_INDEX
        self.BIOTHING_TYPES = list(self.ES_INDICES.keys())

        IOLoop.current().add_callback(self._initialize)

    async def _initialize(self):

        # failures will be logged concisely
        logging.getLogger('elasticsearch.trace').propagate = False

        if should_log_es_py_ver():
            self.logger.info("Python Elasticsearch Version: %s", elasticsearch.__version__)
            self.logger.info("Python Elasticsearch DSL Version: %s", elasticsearch_dsl.__version__)
            if elasticsearch.__version__[0] != elasticsearch_dsl.__version__[0]:
                self.logger.error("ES Pacakge Version Mismatch with ES-DSL.")
        if should_log_es_host_ver(self.ES_HOST):
            versions = await get_es_versions(self.async_es_client)
            self.logger.info('Elasticsearch Version: %s', versions["elasticsearch_version"])
            self.logger.info('Elasticsearch Cluster: %s', versions["elasticsearch_cluster"])
            major_version = versions["elasticsearch_version"].split('.')[0]
            if major_version.isdigit() and int(major_version) != elasticsearch.__version__[0]:
                self.logger.error("ES Python Version Mismatch.")

        # populate source mappings
        for biothing_type in self.ES_INDICES:
            await self.read_index_mappings(biothing_type)

        # resume normal log flow
        logging.getLogger('elasticsearch.trace').propagate = True

    def validate(self):
        '''
        Additional ES settings to validate.
        '''
        super().validate()

        assert isinstance(self.ES_INDEX, str)
        assert isinstance(self.ES_DOC_TYPE, str)
        assert isinstance(self.ES_INDICES, dict)
        assert '*' not in self.ES_DOC_TYPE

    def get_es_client(self):
        '''
        Return the default blocking elasticsearch client.
        The connection is created upon first call.
        '''
        return self._connections.get_connection('sync')

    def get_async_es_client(self):
        '''
        Return the async elasitcsearch client.
        The connection is created upon first call.
        API calls return awaitable objects.
        '''
        return self._connections.get_connection('async')

    async def read_index_mappings(self, biothing_type=None):
        """
        Read ES index mappings for the corresponding biothing_type,
        Populate datasource info and field properties from mappings.
        Return ES raw response. This implementation combines indices.

        The ES response would look like: (for es7+)
        {
            'index_1': {
                'properties': { ... },  ---> source_properties
                '_meta': {
                    "src" : { ... }     ---> source_licenses
                    ...
                },              -----------> source_metadata
                ...
            },
            'index_2': {
                ...     ---------> Combine with results above
            }
        }
        """

        biothing_type = biothing_type or self.ES_DOC_TYPE
        try:
            mappings = await self.async_es_client.indices.get_mapping(
                index=self.ES_INDICES[biothing_type],
                allow_no_indices=True,
                ignore_unavailable=True,
                local=False)
        except elasticsearch.TransportError as exc:
            self.logger.error('Error loading index mapping for [%s].', biothing_type)
            self.logger.debug(str(exc))
            return None

        metadata = self.source_metadata[biothing_type]
        properties = self.source_properties[biothing_type]
        licenses = self.result_transform.source_licenses[biothing_type]

        metadata.clear()
        properties.clear()
        licenses.clear()

        metadata['_biothing'] = biothing_type
        metadata['_indices'] = list(mappings.keys())

        for index in mappings:

            mapping = mappings[index]['mappings']

            if mapping and elasticsearch.__version__[0] < 7:
                # remove doc_type, support 1 type per index
                mapping = next(iter(mapping.values()))

            if '_meta' in mapping:
                for key, val in mapping['_meta'].items():
                    # combine dict from multiple index
                    if key in metadata and isinstance(val, dict) \
                            and isinstance(metadata[key], dict):
                        metadata[key].update(val)
                    else:  # otherwise set/replace
                        metadata[key] = val

                # metadata.update(mapping['_meta'])  # alternative, no combine

                if 'src' in mapping['_meta']:
                    for src, info in mapping['_meta']['src'].items():
                        if 'license_url_short' in info:
                            licenses[src] = info['license_url_short']
                        elif 'license_url' in info:
                            licenses[src] = info['license_url']

            if 'properties' in mapping:
                properties.update(mapping['properties'])

        return mappings

    def get_field_notes(self):
        '''
        Return the cached field notes associated with this instance.
        '''
        return self._fields_notes

    @property
    def es_client(self):
        return self.get_es_client()

    @property
    def async_es_client(self):
        return self.get_async_es_client()



import asyncio

from biothings.utils.hub_db import get_data_plugin
import biothings.hub.dataload.dumper as dumper


class GitDataPlugin(dumper.GitDumper):

    # override to point to "data_plugin" collection instead of src_dump
    # so we don't mix data sources and plugins
    def prepare_src_dump(self):
        self.src_dump = get_data_plugin()
        self.src_doc = self.src_dump.find_one({'_id': self.src_name}) or {}


class ManualDataPlugin(dumper.ManualDumper):

    # override to point to "data_plugin" collection instead of src_dump
    # so we don't mix data sources and plugins
    def prepare_src_dump(self):
        self.src_dump = get_data_plugin()
        self.src_doc = self.src_dump.find_one({'_id': self.src_name}) or {}

    @asyncio.coroutine
    def dump(self, *args, **kwargs):
        yield from super(ManualDataPlugin, self).dump(
            path="",  # it's the version is original method implemention
            # but no version here available
            release="", *args, **kwargs)


class DataPluginManager(dumper.DumperManager):

    def load(self, plugin_name, *args, **kwargs):
        return super(DataPluginManager, self).dump_src(plugin_name, *args, **kwargs)



''' Functions to return versions of things. '''
import functools
import logging
# import pip
import os
import re
import shlex
import sys
from subprocess import DEVNULL, check_output

from git import (Git, GitCommandError, InvalidGitRepositoryError,
                 NoSuchPathError, Repo)

import biothings
from biothings.utils.dataload import dict_sweep


def get_python_version():
    ''' Get a list of python packages installed and their versions. '''
    try:
        output = check_output('pip list', shell=True, stderr=DEVNULL)
        return output.decode('utf-8').replace('\r', '').split('\n')[2: -1]
    except Exception:
        return []

@functools.lru_cache()
def get_biothings_commit():
    ''' Gets the biothings commit information. '''
    try:
        with open(os.path.join(os.path.dirname(biothings.__file__), '.git-info'), 'r') as f:
            lines = [l.strip('\n') for l in f.readlines()]
            return {
                'repository-url': lines[0],
                'commit-hash': lines[1],
                'master-commits': lines[2],
                'version': biothings.get_version()
            }
    except Exception:
        return {
            'repository-url': '',
            'commit-hash': '',
            'master-commits': '',
            'version': biothings.get_version()
        }


@functools.lru_cache()
def get_repository_information(app_dir=None):
    """
    Get the repository information for the local repository, if it exists.
    """
    commit_hash = ''
    repository_url = ''

    if app_dir:
        app_dir = os.path.abspath(app_dir)

        try:
            args = shlex.split("git rev-parse HEAD")
            output = check_output(args, cwd=app_dir, shell=True, stderr=DEVNULL)
            commit_hash = output.decode('utf-8').strip('\n')
        except BaseException:
            pass

        try:
            args = shlex.split("git config --get remote.origin.url")
            output = check_output(args, cwd=app_dir, shell=True, stderr=DEVNULL)
            repository_url = output.decode('utf-8').strip('\n')
        except BaseException:
            pass

    codebase = {
        'repository-url': repository_url,
        'commit-hash': commit_hash
    }
    return codebase


def get_python_exec_version():
    return {
        'version': sys.version,
        'version_info': {
            "major": sys.version_info[0],
            "minor": sys.version_info[1],
            "micro": sys.version_info[2]
        }
    }

@functools.lru_cache()
def get_software_info(app_dir=None):
    return {
        'python-package-info': get_python_version(),
        'codebase': get_repository_information(app_dir=app_dir),
        'biothings': get_biothings_commit(),
        'python-info': get_python_exec_version(),
    }


def check_new_version(folder, max_commits=10):
    """
    Given a folder pointing to a Git repo, return a dict containing info
    about remote commits not qpplied yet to the repo, or empty dict if nothing
    new.
    """
    # from https://stackoverflow.com/questions/8290233/gitpython-get-list-of-remote-commits-not-yet-applied
    repo = Repo(folder)
    try:
        repo_url = re.sub(r"\.git$", "", repo.remotes.origin.url)
    except Exception as e:
        logging.debug("Can't determine repository URL: %s" % e)
        repo_url = None
    new_info = {}
    try:
        # we can't directly get the list of new commits without fetching them locally first
        # but we'd like to avoid fetching all the time just to check.
        # what we can do is a ls-remote and check the HEAD hash, if different, then fetch
        # (no pull) and inspect differences.
        head = repo.head.ref
        tracking = head.tracking_branch()
        # inspect remote HEAD for that branch
        output = repo.git.ls_remote("--heads", tracking.remote_name, tracking.remote_head)
        remote_head_hexsha = output.split("\t")[0]
        if remote_head_hexsha == head.commit.hexsha:
            # hashes the same, we're up-to-date with the remote
            return
        else:
            logging.info("HEAD on remote is different, new commit(s) available for '%s'" % folder)
            logging.info("HEAD(remote): %s, HEAD(local): %s" % (remote_head_hexsha, head.commit.hexsha))
            # need to fetch new code locally
            # usually one remotes, but just in case...
            for remote in repo.remotes:
                remote.fetch()
        # now identify new commits
        new_commits = [commit for commit in tracking.commit.iter_items(repo, f'{head.path}..{tracking.path}')]
        if new_commits:
            new_info = {
                "latest": new_commits[0].hexsha[:6],
                "commits": [
                    {
                        "hash": c.hexsha[:6],
                        "url": repo_url and os.path.join(repo_url, "commit", c.hexsha) or None,
                        "date": c.committed_datetime.isoformat(),
                        "message": c.message
                    } for c in new_commits][:max_commits],
                "total": len(new_commits),
            }
    except Exception as e:

        logging.warning("Can't check for new version: %s" % e)
        raise e

    return new_info

def get_version(folder):
    repo = Repo(folder)  # app or lib dir
    url = repo.remotes.origin.url
    try:
        commit = repo.head.object.hexsha[:6]
        commitdate = repo.head.object.committed_datetime.isoformat()
    except Exception as e:
        logging.warning("can't determine app commit hash: %s" % e)
        commit = "unknown"
        commitdate = "unknown"

    try:
        return {"branch": repo.active_branch.name,
                "commit": commit,
                "date": commitdate,
                "giturl": url}
    except Exception as e:
        logging.warning("can't determine app version, assuming HEAD detached': %s" % e)
        return {"branch": "HEAD detached",
                "commit": commit,
                "date": commitdate,
                "giturl": url}


def set_versions(config, app_folder):
    """
    Propagate versions (git branch name) in config module.
    Also set app and biothings folder paths (though not
    exposed as a config param since they are lower-cased,
    see biothings.__init__.py, regex PARAM_PAT)
    """
    if not os.path.exists(app_folder):
        raise FileNotFoundError("'%s' application folder doesn't exist")
    # app_version: version of the API application
    if not hasattr(config, "APP_VERSION"):
        config.APP_VERSION = get_version(app_folder)
        config.app_folder = app_folder
    else:
        logging.info("app_version '%s' forced in configuration file" % config.APP_VERSION)

    # biothings_version: version of BioThings SDK
    if not hasattr(config, "BIOTHINGS_VERSION"):
        import biothings
        # .../biothings.api/biothings/__init__.py
        bt_folder, _bt = os.path.split(os.path.split(os.path.realpath(biothings.__file__))[0])
        if not os.path.exists(bt_folder):
            raise FileNotFoundError("'%s' biothings folder doesn't exist")
        assert _bt == "biothings", "Expectig 'biothings' dir in biothings lib path"
        config.BIOTHINGS_VERSION = get_version(bt_folder)
        config.biothings_folder = bt_folder
    else:
        logging.info("biothings_version '%s' forced in configuration file" %
                     config.BIOTHINGS_VERSION)

    logging.info("Running app_version=%s with biothings_version=%s" %
                 (repr(config.APP_VERSION), repr(config.BIOTHINGS_VERSION)))


def get_source_code_info(src_file):
    """
    Given a path to a source code, try to find information
    about repository, revision, URL pointing to that file, etc...
    Return None if nothing can be determined.
    Tricky cases:
      - src_file could refer to another repo, within current repo
        (namely a remote data plugin, cloned within the api's plugins folder
      - src_file could point to a folder, when for instance a dataplugin is
        analized. This is because we can't point to an uploader file since
        it's dynamically generated
    """
    # need to be absolute to build proper github URL
    abs_src_file = os.path.abspath(src_file)
    try:
        repo = Repo(abs_src_file, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        logging.exception("Can't find a github repository for file '%s'" % src_file)
        return None
    try:
        gcmd = Git(repo)
        hash = gcmd.rev_list(-1, repo.active_branch, abs_src_file)
        rel_src_file = abs_src_file.replace(repo.working_dir, "").strip("/")
        if not hash:
            # seems to be a repo cloned within a repo, change directory
            curdir = os.path.abspath(os.curdir)
            try:
                if os.path.isdir(abs_src_file):
                    os.chdir(abs_src_file)
                    hash = gcmd.rev_list(-1, repo.active_branch)
                else:
                    dirname, filename = os.path.split(abs_src_file)
                    os.chdir(dirname)
                    hash = gcmd.rev_list(-1, repo.active_branch, filename)
                rel_src_file = ""  # will point to folder by commit hash
            finally:
                os.chdir(curdir)
        if hash:
            short_hash = gcmd.rev_parse(hash, short=7)
        else:
            logging.warning("Couldn't determine commit hash for file '%s'" % src_file)
            hash = None
            short_hash = None
        # could have more than one URLs for origin, only take first
        repo_url = next(repo.remote().urls)
        info = {
            "repo": repo_url,
            "commit": short_hash,
            "branch": repo.active_branch.name,
        }
        if os.path.isdir(src_file):
            info["folder"] = rel_src_file
        else:
            info["file"] = rel_src_file
        info = dict_sweep(info)
        # rebuild URL to that file
        if "github.com" in repo_url:
            info["url"] = os.path.join(re.sub(r"\.git$", "", repo_url),
                                       "tree", hash, rel_src_file)

        return info

    except GitCommandError:
        logging.exception("Error while getting git information for file '%s'" % src_file)
        return None
    except TypeError as e:
        # happens with biothings symlink, just ignore
        logging.debug("Can't determine source code info (but that's fine): %s" % e)
        return None



"""
    Biothings Web API

    Run API:

    1a. with default settings:

        # serve all data on localhost:9200
        # may be used for testing and development

        from biothings.web import BiothingsAPI
        api = BiothingsAPI()
        api.start()

    1b. with a customized config module:

        # allow detailed configuration of apis
        # this is similar to how the hub starts web apis

        from biothings.web import BiothingsAPI
        import config
        api = BiothingsAPI(config)
        api.start()

    2. with default command line options:

        # by defualt looks for a 'config.py' under cwd
        # similar to discovery and crawler biothings apps

        from biothings.web.index_base import main
        if __name__ == '__main__':
            main()

    3. with application template framework. # TODO

    On top of the common ways described above, you can specify
    the customized config module in the following ways:

    1. a python module already imported

    2. a fully qualified name to import, for example:
     - 'config'
     - 'app.config.dev'

    3. a file path to a python module, for example:
     - '../config.py'
     - '/home/ubuntu/mygene/config.py'
     - 'C:\\Users\\Biothings\\mygene\\config.py'

    4. explicitly specify None or '' to use default.

    See below for additional configurations like
    using an external asyncio event loop.
"""

import logging

import tornado.httpserver
import tornado.ioloop
import tornado.web

from biothings import get_version
from biothings.web.settings import BiothingESWebSettings


class BiothingsAPI():
    """
    Configure a Biothings Web API Server.

    There are three parts to it:
    * A biothings config module that defines the API handlers.
    * Additional Tornado handlers and application settings.
    * An asyncio event loop to run the tornado application.

    The API can be started with:
    * An external event loop by calling get_server()
    * A default tornado event loop by calling start()

    Unless started externally, debug mode:
    * Sets proper logging levels for root logger and es,
    * Enables debug mode on tornado except for autoreload,
    * Disables integrated tracking and error monitoring.
    """

    def __init__(self, config_module=None):
        # About debug mode in tornado:
        # https://www.tornadoweb.org/en/stable/guide/running.html \
        # #debug-mode-and-automatic-reloading
        logging.info("Biothings API %s", get_version())
        self.config = BiothingESWebSettings(config_module)
        self.handlers = []  # additional handlers
        self.settings = dict(debug=False)
        self.host = None

    @staticmethod
    def use_curl():
        """
        Use curl implementation for tornado http clients.
        More on https://www.tornadoweb.org/en/stable/httpclient.html
        """
        tornado.httpclient.AsyncHTTPClient.configure(
            "tornado.curl_httpclient.CurlAsyncHTTPClient")

    def update(self, **settings):
        """
        Update Tornado application settings. More on:
        https://www.tornadoweb.org/en/stable/web.html \
        #tornado.web.Application.settings
        """
        self.settings.update(settings)

    def _configure_logging(self):
        root_logger = logging.getLogger()
        self.config.configure_logger(root_logger)
        logging.getLogger('urllib3').setLevel(logging.ERROR)
        logging.getLogger('elasticsearch').setLevel(logging.WARNING)
        if self.settings['debug']:
            root_logger.setLevel(logging.DEBUG)
            es_tracer = logging.getLogger('elasticsearch.trace')
            es_tracer.setLevel(logging.DEBUG)
            es_tracer.addHandler(logging.NullHandler())
        else:
            root_logger.setLevel(logging.INFO)

    def get_server(self):
        """
        Run API in an external event loop.
        """
        webapp = self.config.get_app(self.settings, self.handlers)
        server = tornado.httpserver.HTTPServer(webapp, xheaders=True)
        return server

    def start(self, port=8000):
        """
        Run API in the default event loop.
        """
        self._configure_logging()

        http_server = self.get_server()
        http_server.listen(port, self.host)

        logger = logging.getLogger('biothings.web')
        logger.info('Server is running on "%s:%s"...',
                    self.host or '0.0.0.0', port)

        loop = tornado.ioloop.IOLoop.instance()
        loop.start()


BiothingsAPIApp = BiothingsAPI



'''
Backend for storing merged genedoc after building.
Support MongoDB, ES, CouchDB
'''
import os
from functools import partial
from biothings.utils.common import get_timestamp, get_random_string
from biothings.utils.backend import DocBackendBase, DocMongoBackend, DocESBackend
from biothings.utils.es import ESIndexer
import biothings.utils.mongo as mongo
from biothings.utils.hub_db import get_source_fullname, get_src_build


# Source specific backend (deals with build config, master docs, etc...)
class SourceDocBackendBase(DocBackendBase):
    def __init__(self, build_config, build, master, dump, sources):
        if type(build_config) == partial:
            self._build_config_provider = build_config
            self._build_config = None
        else:
            self._build_config_provider = None
            self._build_config = build_config
        if type(build) == partial:
            self._build_provider = build
            self._build = None
        else:
            self._build_provider = None
            self._build = build
        if type(master) == partial:
            self._master_provider = master
            self._master = None
        else:
            self._master_provider = None
            self._master = master
        if type(dump) == partial:
            self._dump_provider = dump
            self._dump = None
        else:
            self._dump_provider = None
            self._dump = dump
        if type(sources) == partial:
            self._sources_provider = sources
            self._sources = None
        else:
            self._sources_provider = None
            self._sources = sources
        self._build_config = None
        self.src_masterdocs = None
        # keep track of sources which were accessed
        self.sources_accessed = {}

    def __getattr__(self, attr):
        if attr not in ["build_config", "build", "dump", "master", "sources"]:
            return AttributeError(attr)
        privattr = "_" + attr
        if getattr(self, privattr) is None:
            if getattr(self, privattr + "_provider"):
                res = getattr(self, privattr + "_provider")()
            else:
                res = getattr(self, privattr)
            setattr(self, privattr, res)
        return getattr(self, privattr)

    def get_build_configuration(self, build_name):
        raise NotImplementedError("sub-class and implement me")

    def get_src_master_docs(self):
        raise NotImplementedError("sub-class and implement me")

    def validate_sources(self, sources=None):
        raise NotImplementedError("sub-class and implement me")

    def get_src_metadata(self):
        raise NotImplementedError("sub-class and implement me")

    def __getitem__(self, src_name):
        self.sources_accessed[src_name] = 1
        return self.sources[src_name]


# Target specific backend
class TargetDocBackend(DocBackendBase):
    def __init__(self, *args, **kwargs):
        super(TargetDocBackend, self).__init__(*args, **kwargs)
        self._target_name = None

    def set_target_name(self, target_name, build_name=None):
        """
        Create/prepare a target backend, either strictly named "target_name"
        or named derived from "build_name" (for temporary backends)
        """
        self._target_name = target_name or self.generate_target_name(
            build_name)

    @property
    def target_name(self):
        return self._target_name

    def generate_target_name(self, build_config_name):
        assert build_config_name is not None
        return '{}_{}_{}'.format(build_config_name, get_timestamp(),
                                 get_random_string()).lower()

    def get_backend_url(self):
        """
        Return backend URL (see create_backend() for formats)
        """
        # default is a collection in target database
        return self._target_name

    def post_merge(self):
        pass


class SourceDocMongoBackend(SourceDocBackendBase):
    def get_build_configuration(self, build_name):
        self._config = self.build_config.find_one({'_id': build_name})
        return self._config

    def validate_sources(self, sources=None):
        assert self._build_config, "'self._build_config' cannot be empty."

    def get_src_master_docs(self):
        if self.src_masterdocs is None:
            self.src_masterdocs = dict([(src['_id'], src)
                                        for src in list(self.master.find())])
        return self.src_masterdocs

    def get_src_metadata(self):
        """
        Return source versions which have been previously accessed wit this backend object
        or all source versions if none were accessed. Accessing means going through __getitem__
        (the usual way) and allows to auto-keep track of sources of interest, thus returning
        versions only for those.
        """
        src_version = {}
        # what's registered in each uploader, from src_master.
        # also includes versions
        src_meta = {}
        srcs = []
        if self.sources_accessed:
            for src in self.sources_accessed:
                fullname = get_source_fullname(src)
                main_name = fullname.split(".")[0]
                doc = self.dump.find_one({"_id": main_name})
                srcs.append(doc["_id"])
            srcs = list(set(srcs))
        else:
            srcs = [d["_id"] for d in self.dump.find()]
        # we need to return main_source named, but if accessed, it's been through sub-source names
        # query is different in that case
        for src in self.dump.find({"_id": {"$in": srcs}}):
            # now merge other extra information from src_master (src_meta key). src_master _id
            # are sub-source names, not main source so we have to deal with src_dump as well
            # in order to resolve/map main/sub source name
            subsrc_versions = []
            if src and src.get("upload"):
                meta = {}
                for job_name in src["upload"].get("jobs", {}):
                    job = src["upload"]["jobs"][job_name]
                    # "step" is the actual sub-source name
                    docm = self.master.find_one({"_id": job.get("step")})
                    if docm and docm.get("src_meta"):
                        meta[job.get("step")] = docm["src_meta"]
                # when more than 1 sub-sources, we can have different version in sub-sources
                # (not normal) if one sub-source uploaded, then dumper downloaded a new version,
                # then the other sub-source uploaded that version. This should never happen, just make sure
                subsrc_versions = [{"sub-source": job.get("step"), "version": job.get("release")}
                                   for job in src["upload"].get("jobs", {}).values()]
                assert len(set([s["version"] for s in subsrc_versions])) == 1, "Expecting one version " + \
                    "in upload sub-sources for main source '%s' but got: %s" % (src["_id"], subsrc_versions)
                # usually, url & license are the same wathever the sub-sources are. They are
                # share common metadata, and we don't want them to be repeated for each sub-sources.
                # but, code key is always different for instance and must specific for each sub-sources
                # here we make sure to factor common keys, while the specific ones at sub-level
                if len(meta) > 1:
                    common = {}
                    any = list(meta)[0]
                    topop = []  # common keys
                    for anyk in meta[any]:
                        if len({
                                meta[s].get(anyk) == meta[any][anyk]
                                for s in meta
                        }) == 1:
                            topop.append(anyk)
                    for k in topop:
                        common[k] = meta[any][k]
                        [meta[subname].pop(k, None) for subname in meta]

                    for k, v in common.items():
                        src_meta.setdefault(src["_id"], {}).setdefault(k, v)
                    for subname in meta:
                        for k, v in meta[subname].items():
                            src_meta.setdefault(src["_id"], {}).setdefault(
                                k, {}).setdefault(subname, v)
                # we have metadata, but just one (ie. no sub-source), don't display it
                elif meta:
                    assert len(meta) == 1
                    subname, metad = meta.popitem()
                    for k, v in metad.items():
                        src_meta.setdefault(src["_id"], {}).setdefault(k, v)
            if subsrc_versions:
                version = subsrc_versions[0]["version"]
                src_version[src['_id']] = version
                src_meta.setdefault(src["_id"],
                                    {}).setdefault("version", version)
        return src_meta


class TargetDocMongoBackend(TargetDocBackend, DocMongoBackend):
    def set_target_name(self, target_name=None, build_name=None):
        super(TargetDocMongoBackend,
              self).set_target_name(target_name, build_name)
        self.target_collection = self.target_db[self._target_name]


class ShardedTargetDocMongoBackend(TargetDocMongoBackend):
    def prepare(self):
        assert self.target_name, "target_name not set"
        assert self.target_db, "target_db not set"
        dba = self.target_db.client.admin
        dba.command("shardCollection",
                    "%s.%s" % (self.target_db.name, self.target_name),
                    key={"_id": "hashed"})


class LinkTargetDocMongoBackend(TargetDocBackend):
    """
    This backend type act as a dummy target backend, the data
    is actually stored in source database. It means only one
    datasource can be linked to that target backend, as a consequence,
    when this backend is used in a merge, there's no actual data
    merge. This is useful when "merging/indexing" only one datasource,
    where the merge step is just a duplication of datasource data.
    """
    name = 'link'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # will be set later by LinkDataBuilder, should be the name
        # of the datasource in src database
        self.datasource_name = None
        self.source_db = None

    @property
    def target_collection(self):
        assert self.source_db
        assert self.datasource_name
        return self.source_db[self.datasource_name]

    def get_backend_url(self):
        assert self.datasource_name
        return ("src", self.datasource_name)

    def drop(self):
        # nothing to drop we don't store data
        # but needs to be implemented
        pass


def create_backend(db_col_names, name_only=False, follow_ref=False, **kwargs):
    """
    Guess what's inside 'db_col_names' and return the corresponding backend.
    - It could be a string (will first check for an src_build doc to check
      a backend_url field, if nothing there, will lookup a mongo collection
      in target database)
    - or a tuple("target|src","col_name")
    - or a ("mongodb://user:pass@host","db","col_name") URI.
    - or a ("es_host:port","index_name","doc_type")
    If name_only is true, just return the name uniquely identifying the collection or index
    URI connection.
    """
    col = None
    db = None
    is_mongo = True
    if type(db_col_names) == str:
        # first check build doc, if there's backend_url key, we'll use it instead of
        # direclty using db_col_names as target collection (see LinkDataBuilder)
        bdoc = get_src_build().find_one({"_id": db_col_names})
        if follow_ref and bdoc and bdoc.get(
                "backend_url") and bdoc["backend_url"] != db_col_names:
            return create_backend(bdoc["backend_url"],
                                  name_only=name_only,
                                  follow_ref=follow_ref,
                                  **kwargs)
        else:
            db = mongo.get_target_db()
            col = db[db_col_names]
            # normalize params
            db_col_names = [
                "%s:%s" % (db.client.HOST, db.client.PORT), db.name, col.name
            ]
    elif db_col_names[0].startswith("mongodb://"):
        assert len(
            db_col_names
        ) == 3, "Missing connection information for %s" % repr(db_col_names)
        conn = mongo.MongoClient(db_col_names[0])
        db = conn[db_col_names[1]]
        col = db[db_col_names[2]]
        # normalize params
        db_col_names = [
            "%s:%s" % (db.client.HOST, db.client.PORT), db.name, col.name
        ]
    elif len(db_col_names) == 3 and ":" in db_col_names[0]:
        is_mongo = False
        idxr = ESIndexer(index=db_col_names[1],
                         doc_type=db_col_names[2],
                         es_host=db_col_names[0],
                         **kwargs)
        db = idxr
        col = db_col_names[1]
    else:
        assert len(
            db_col_names
        ) == 2, "Missing connection information for %s" % repr(db_col_names)
        db = db_col_names[0] == "target" and mongo.get_target_db(
        ) or mongo.get_src_db()
        col = db[db_col_names[1]]
        # normalize params (0:host, 1:port)
        db_col_names = [
            "%s:%s" % (db.client.address[0], db.client.address[1]), db.name,
            col.name
        ]
    assert col is not None, "Could not create collection object from %s" % repr(
        db_col_names)
    if name_only:
        if is_mongo:
            return "mongo_%s_%s_%s" % (db_col_names[0].replace(
                ":", "_"), db_col_names[1], db_col_names[2])
        else:
            return "es_%s_%s_%s" % (db_col_names[0].replace(
                ":", "_"), db_col_names[1], db_col_names[2])
    else:
        if is_mongo:
            return DocMongoBackend(db, col)
        else:
            return DocESBackend(db)


def generate_folder(root_folder, old_db_col_names, new_db_col_names):
    new = create_backend(new_db_col_names, name_only=True)
    old = create_backend(old_db_col_names, name_only=True)
    diff_folder = os.path.join(root_folder, "%s-%s" % (old, new))
    return diff_folder


def merge_src_build_metadata(build_docs):
    """
    Merge metadata from src_build documents. A list of docs
    should be passed, the order is important: the 1st element
    has the less precedence, the last the most. It means that,
    when needed, some values from documents on the "left" of the
    list may be overridden by one on the right.
    Ex: build_version field
    Ideally, build docs shouldn't have any sources in common to
    prevent any unexpected conflicts...
    """
    assert type(build_docs) == list and len(build_docs) >= 2, \
        "More than one build document must be passed in order metadata"
    first = build_docs[0]
    others = build_docs[1:]
    meta = first.get("_meta", {})
    for new_doc in others:
        new_meta = new_doc.get("_meta", {})
        for k, v in new_meta.items():
            # src_version, src, stats: merge
            if type(v) == dict:
                meta.setdefault(k, {})
                meta[k].update(v)
            # build_version, build_date: override
            else:
                meta[k] = v
    # TODO: some fields in stats don't make when merged: total, observed/vcf in mv
    return meta



import json


def make_object(attr, value):
    '''
    Create dictionary following the input dot notation and the value
    Example::

        make_object('a.b.c', 100) -->
        or make_object(['a','b','c'], 100) -->
        {a:{b:{c:100}}}
    '''
    attr_list = attr.split('.')
    s = ''
    for k in attr_list:
        s += '{"' + k + '":'
    s += json.dumps(value)
    s += "}" * (len(attr_list))
    return json.loads(s)


def merge_object(obj1, obj2):
    for k in obj2:
        try:
            if isinstance(obj2[k], dict):
                obj1[k] = merge_object(obj1[k], obj2[k])
            else:
                obj1[k] = obj2[k]
        except:
            obj1[k] = obj2[k]
    return obj1


def parse_dot_fields(genedoc):
    """
    parse_dot_fields({'a': 1, 'b.c': 2, 'b.a.c': 3})
     should return
        {'a': 1, 'b': {'a': {'c': 3}, 'c': 2}}
    """
    dot_fields = []
    expanded_doc = {}
    for key in genedoc:
        if key.find('.') != -1:
            dot_fields.append(key)
            expanded_doc = merge_object(expanded_doc, make_object(key, genedoc[key]))
    genedoc.update(expanded_doc)
    for key in dot_fields:
        del genedoc[key]
    return genedoc


def compose_dot_fields_by_fields(genedoc, fields):
    """
    reverse funtion of parse_dot_fields
    """
    res = None
    to_del = set()
    for k in fields:
        if k.find('.') != -1:
            if not res:
                import copy
                res = copy.deepcopy(genedoc)
            ks = k.split('.')
            broke = False
            if ks[0] in genedoc:
                t = genedoc[ks[0]]
                for e in ks[1:]:
                    if e in t:
                        t = t[e]
                    else:
                        broke = True
                        break
                if broke:
                    continue
                to_del.add(ks[0])
                res[k] = t

    for k in to_del:
        del res[k]

    return res if res else genedoc



import biothings.utils.redis as redis

class IDCache(object):

    def mark_done(self, _ids):
        raise NotImplementedError()

    def load(self, name, id_provider, flush=True):
        """
        name is the cache name
        id_provider returns batch of ids, ie. list(_ids)
        flush to delete existing cache
        """
        raise NotImplementedError()


class RedisIDCache(IDCache):

    def __init__(self, name, connection_params):
        self.name = name
        self.redis_client = redis.RedisClient(connection_params)
        try:
            self.redis_client.check()
        except redis.RedisClientError:
            self.redis_client.initialize()
            self.redis_client.check()

    def load(self, id_provider, flush=True):
        db = self.redis_client.get_db(self.name)
        if flush:
            db.flushdb()
        for _ids in id_provider:
            dids = dict([(_id, 0) for _id in _ids])
            db.mset(dids)

    def mark_done(self, _ids):
        db = self.redis_client.get_db(self.name)
        db.delete(*_ids)



"""
    Elasticsearch Query Execution
"""
import asyncio

from elasticsearch import (ConnectionError, NotFoundError, RequestError,
                           TransportError)
from tornado.web import HTTPError

from biothings.web.handlers.exceptions import BadRequest, EndRequest


class ESQueryBackend(object):
    '''
    Execute an Elasticsearch query
    '''

    def __init__(self, web_settings):

        self.client = web_settings.get_async_es_client()

        # es indices
        self.indices = web_settings.ES_INDICES
        self.default_index = web_settings.ES_INDEX
        self.default_type = web_settings.ES_DOC_TYPE

        # for scroll queries
        self.scroll_time = web_settings.ES_SCROLL_TIME
        self.scroll_size = web_settings.ES_SCROLL_SIZE

    async def execute(self, query, options):
        '''
        Execute the corresponding query.
        May override to add more. Handle uncaught exceptions.
        Must return an awaitable.
        '''
        if options.scroll_id:
            try:
                res = await self.client.scroll(
                    scroll_id=options.scroll_id,
                    scroll=self.scroll_time)
            except ConnectionError:
                raise HTTPError(503)
            except (NotFoundError, RequestError, TransportError):
                raise BadRequest(reason="Invalid or stale scroll_id.")
            else:
                if not res['hits']['hits']:
                    raise EndRequest(reason="No more results to return.")
                return res

        if query:
            biothing_type = options.pop('biothing_type', None) or self.default_type
            query = query.index(self.indices.get(biothing_type, self.default_index))

            if options.pop('fetch_all', False):
                query = query.params(scroll=self.scroll_time)
                query = query.extra(size=self.scroll_size)
            try:
                res = await query.using(self.client).execute()
            except ConnectionError:
                raise HTTPError(503)
            except RequestError as exc:
                raise BadRequest(_es_error=exc)
            except TransportError as exc:
                if exc.error == 'search_phase_execution_exception':
                    raise BadRequest(reason=exc.error)
                elif exc.error == 'index_not_found_exception':
                    raise HTTPError(500, reason=exc.error)
                elif exc.status_code == 'N/A':
                    raise HTTPError(503)
                else:  # unexpected
                    raise
            else:  # format to {} or [{}...]
                if isinstance(res, list):
                    return [res_.to_dict() for res_ in res]
                return res.to_dict()

        return asyncio.sleep(0, {})



import sys
import os
import subprocess

from biothings.utils.hub_db import get_src_dump, get_data_plugin, get_hub_db_conn, backup, restore
from biothings import config
from biothings.hub.dataload.dumper import GitDumper

logging = config.logger


def migrate_0dot1_to_0dot2():
    """
    mongodb src_dump/data_plugin changed:
        1. "data_folder" and "release" under "download"
        2. "data_folder" and "release" in upload.jobs[subsrc] taken from "download"
        3. no more "err" under "upload"
        4. no more "status" under "upload"
        5. "pending_to_upload" is now "pending": ["upload"]
    """
    src_dump = get_src_dump()
    data_plugin = get_data_plugin()
    for srccol in [src_dump, data_plugin]:
        logging.info("Converting collection %s" % srccol)
        srcs = [src for src in srccol.find()]
        wasdue = False
        for src in srcs:
            logging.info("\tConverting '%s'" % src["_id"])
            # 1.
            for field in ["data_folder", "release"]:
                if field in src:
                    logging.debug(
                        "\t\t%s: found '%s' in document, moving under 'download'"
                        % (src["_id"], field))
                    try:
                        src["download"][field] = src.pop(field)
                        wasdue = True
                    except KeyError as e:
                        logging.warning(
                            "\t\t%s: no such field '%s' found, skip it (error: %s)"
                            % (src["_id"], field, e))
            # 2.
            for subsrc_name in src.get("upload", {}).get("jobs", {}):
                for field in ["data_folder", "release"]:
                    if field not in src["upload"]["jobs"][subsrc_name]:
                        logging.debug(
                            "\t\t%s: no '%s' found in upload jobs, taking it from 'download' (or from root keys)"
                            % (src["_id"], field))
                        try:
                            src["upload"]["jobs"][subsrc_name][field] = src[
                                "download"][field]
                            wasdue = True
                        except KeyError:
                            try:
                                src["upload"]["jobs"][subsrc_name][
                                    field] = src[field]
                                wasdue = True
                            except KeyError:
                                logging.warning(
                                    "\t\t%s: no such field '%s' found, skip it"
                                    % (src["_id"], field))
            # 3. & 4.
            for field in ["err", "status"]:
                if field in src.get("upload", {}):
                    logging.debug("\t\t%s: removing '%s' key from 'upload'" %
                                  (src["_id"], field))
                    src["upload"].pop(field)
                    wasdue = True
            # 5.
            if "pending_to_upload" in src:
                logging.debug(
                    "\t%s: found 'pending_to_upload' field, moving to 'pending' list"
                    % src["_id"])
                src.pop("pending_to_upload")
                wasdue = True
                if "upload" not in src.get("pending", []):
                    src.setdefault("pending", []).append("upload")
            if wasdue:
                logging.info("\tFinishing converting document for '%s'" %
                             src["_id"])
                srccol.save(src)
            else:
                logging.info("\tDocument for '%s' already converted" %
                             src["_id"])


def migrate(from_version, to_version, restore_if_failure=True):
    func_name = "migrate_%s_to_%s" % (from_version.replace(
        ".", "dot"), to_version.replace(".", "dot"))
    # backup
    db = get_hub_db_conn()[config.DATA_HUB_DB_DATABASE]
    logging.info("Backing up %s" % db.name)
    path = backup()
    logging.info("Backup file: %s" % path)
    thismodule = sys.modules[__name__]
    try:
        func = getattr(thismodule, func_name)
    except AttributeError:
        logging.error(
            "Can't upgrade, no such function to migrate from '%s' to '%s'" %
            (from_version, to_version))
        raise
    # resolve A->C = A->B then B->C
    logging.info("Start upgrading from '%s' to '%s'" %
                 (from_version, to_version))
    try:
        func()
    except Exception as e:
        logging.exception("Failed upgrading: %s" % e)
        if restore_if_failure:
            logging.info("Now restoring original database from '%s" % path)
            restore(db, path, drop=True)
            logging.info(
                "Done. If you want to keep converted data for inspection, use restore_if_failure=False"
            )
        else:
            logging.info(
                "*not* restoring original data. It can still be restored using file '%s'"
                % path)


class BaseSystemUpgrade(GitDumper):

    def _pull(self, *args):
        old = os.path.abspath(os.curdir)
        try:
            # SRC_ROOT_FOLDER is set to code base root
            os.chdir(self.__class__.SRC_ROOT_FOLDER)
            cmd = ["git", "pull", "--rebase"]
            self.logger.debug("Pulling new code in '%s' using: %s" % (self.__class__.SRC_ROOT_FOLDER, cmd))
            subprocess.check_call(cmd)
        finally:
            os.chdir(old)


class BioThingsSystemUpgrade(BaseSystemUpgrade):
    """Used to upgrade biothings SDK code"""
    SRC_NAME = "__biothings_sdk"  # "__" means private, not listed


class ApplicationSystemUpgrade(BaseSystemUpgrade):
    """Used to upgrade application (api app) code"""
    SRC_NAME = "__application"  # "__" means private, not listed
    



''' Backend access class. '''
from functools import partial

from elasticsearch.exceptions import NotFoundError, TransportError

from biothings.utils.es import ESIndexer
from biothings import config as btconfig


# Generic base backend
class DocBackendBase(object):
    name = 'Undefined'

    def prepare(self):
        '''if needed, add extra preparation steps here.'''

    @property
    def target_name(self):
        raise NotImplementedError()

    @property
    def version(self):
        raise NotImplementedError()

    def insert(self, doc_li):
        raise NotImplementedError()

    def update(self, id, extra_doc):
        '''update only, no upsert.'''
        raise NotImplementedError()

    def drop(self):
        raise NotImplementedError()

    def get_id_list(self):
        raise NotImplementedError()

    def get_from_id(self, id):
        raise NotImplementedError()

    def finalize(self):
        '''if needed, for example for bulk updates, perform flush
           at the end of updating.
           Final optimization or compacting can be done here as well.
        '''

class DocMemoryBackend(DocBackendBase):
    name = 'memory'

    def __init__(self, target_name=None):
        """target_dict is None or a dict."""
        self.target_dict = {}
        self._target_name = target_name or "unnamed"

    @property
    def target_name(self):
        return self._target_name

    def insert(self, doc_li):
        for doc in doc_li:
            self.target_dict[doc['_id']] = doc

    def update(self, id, extra_doc):
        current_doc = self.target_dict.get(id, None)
        if current_doc:
            current_doc.update(extra_doc)
            self.target_dict[id] = current_doc

    def drop(self):
        self.target_dict = {}

    def get_id_list(self):
        return self.target_dict.keys()

    def get_from_id(self, id):
        return self.target_dict[id]

    def finalize(self):
        '''dump target_dict into a file.'''
        from biothings.utils.common import dump
        dump(self.target_dict, self.target_name + '.pyobj')

class DocMongoBackend(DocBackendBase):
    name = 'mongo'

    def __init__(self, target_db, target_collection=None):
        """target_collection is a pymongo collection object."""
        if callable(target_db):
            self._target_db_provider = target_db
            self._target_db = None
        else:
            self._target_db = target_db
        if target_collection:
            self.target_collection = target_collection

    def __eq__(self, other):
        if not isinstance(other, DocMongoBackend):
            return False
        return self.target_name == other.target_name and \
               self.target_collection.database.name == other.target_collection.database.name and \
               self.target_collection.database.client.address == other.target_collection.database.client.address

    @property
    def target_name(self):
        return self.target_collection.name

    @property
    def version(self):
        import biothings.utils.mongo as mongo
        if self.target_collection.database.name == btconfig.DATA_SRC_DATABASE:
            fulln = mongo.get_source_fullname(self.target_collection.name)
            if not fulln:
                return
            mainsrc = fulln.split(".")[0]
            col = mongo.get_src_dump()
            src = col.find_one({"_id": mainsrc})
            return src.get("release")
        elif self.target_collection.database.name == btconfig.DATA_TARGET_DATABASE:
            col = mongo.get_src_build()
            tgt = col.find_one({"_id": self.target_collection.name})
            if not tgt:
                return
            return tgt.get("_meta", {}).get("build_version")
        else:
            return None

    @property
    def target_db(self):
        if self._target_db is None:
            self._target_db = self._target_db_provider()
        return self._target_db

    def count(self):
        return self.target_collection.count()

    def insert(self, docs):
        try:
            res = self.target_collection.insert_many(documents=docs)
            return len(res.inserted_ids)
        except Exception as e:
            import pickle
            pickle.dump(e, open("err", "wb"))

    def update(self, docs, upsert=False):
        '''if id does not exist in the target_collection,
            the update will be ignored except if upsert is True
        '''
        bulk = self.target_collection.initialize_ordered_bulk_op()
        at_least_one = False
        for doc in docs:
            at_least_one = True
            op = bulk.find({'_id': doc["_id"]})
            if upsert:
                op = op.upsert()
            op.update({"$set": doc})
        if at_least_one:
            res = bulk.execute()
            # if doc is the same, it'll be matched but not modified.
            # but for us, it's been processed. if upserted, then it can't be matched
            # before (so matched cound doesn't include upserted). finally, it's only update
            # ops, so don't count nInserted and nRemoved
            return res["nMatched"] + res["nUpserted"]
        else:
            return 0

    def update_diff(self, diff, extra={}):
        '''update a doc based on the diff returned from diff.diff_doc
            "extra" can be passed (as a dictionary) to add common fields to the
            updated doc, e.g. a timestamp.
        '''
        _updates = {}
        _add_d = dict(list(diff.get('add', {}).items()) + list(diff.get('update', {}).items()))
        if _add_d or extra:
            if extra:
                _add_d.update(extra)
            _updates['$set'] = _add_d
        if diff.get('delete', None):
            # _updates['$unset'] = dict([(x, 1) for x in diff['delete']])    # TODO: remove this line, rewritten using dict comprehension
            _updates['$unset'] = {x: 1 for x in diff['delete']}
        res = self.target_collection.update_one({'_id': diff['_id']}, _updates, upsert=False)
        return res.modified_count

    def drop(self):
        self.target_collection.drop()

    def get_id_list(self):
        return [x['_id'] for x in self.target_collection.find(projection=[], manipulate=False)]

    def get_from_id(self, id):
        return self.target_collection.find_one({"_id": id})

    def mget_from_ids(self, ids, asiter=False):
        '''ids is an id list.
           returned doc list should be in the same order of the
             input ids. non-existing ids are ignored.
        '''
        #this does not return doc in the same order of ids
        cur = self.target_collection.find({'_id': {'$in': ids}})
        if asiter:
            return cur
        else:
            return [doc for doc in cur]

    def count_from_ids(self, ids, step=100000):
        '''return the count of docs matching with input ids
           normally, it does not need to query in batches, but MongoDB
           has a BSON size limit of 16M bytes, so too many ids will raise a
           pymongo.errors.DocumentTooLarge error.
        '''
        total_cnt = 0
        for i in range(0, len(ids), step):
            _ids = ids[i:i + step]
            _cnt = self.target_collection.find({'_id': {'$in': _ids}}).count()
            total_cnt += _cnt
        return total_cnt

    def finalize(self):
        '''flush all pending writes.'''
        # no need to flush, fsync is used for backups. also, this locks the whole
        # database for reads...
        #self.target_collection.database.client.fsync(async=True)

    def remove_from_ids(self, ids, step=10000):
        deleted = 0
        for i in range(0, len(ids), step):
            res = self.target_collection.delete_many({'_id': {'$in': ids[i:i + step]}})
            deleted += res.deleted_count
        return deleted

# backward-compatible
DocMongoDBBackend = DocMongoBackend

class DocESBackend(DocBackendBase):
    name = 'es'

    def __init__(self, esidxer=None):
        """esidxer is an instance of utils.es.ESIndexer class."""
        if type(esidxer) == partial:
            self._target_esidxer_provider = esidxer
            self._target_esidxer = None
        else:
            _target_esidxer_provider = None
            self._target_esidxer = esidxer

    @property
    def target_esidxer(self):
        if not self._target_esidxer:
            self._target_esidxer = self._target_esidxer_provider()
        return self._target_esidxer

    @property
    def target_name(self):
        return self.target_esidxer._index

    @property
    def version(self):
        try:
            mapping = self.target_esidxer.get_mapping_meta()
            if mapping.get("_meta"):
                return mapping["_meta"].get("build_version")
        except NotFoundError:
            # index doesn't even exist
            return None

    def prepare(self, update_mapping=True):
        self.target_esidxer.create_index()
        self.target_esidxer.verify_mapping(update_mapping=update_mapping)

    def count(self):
        try:
            return self.target_esidxer.count()
        except TransportError:
            return None

    def insert(self, doc_li):
        self.target_esidxer.add_docs(doc_li)

    def update(self, id, extra_doc):
        self.target_esidxer.update(id, extra_doc, bulk=True)

    def drop(self):
        from utils.es import IndexMissingException

        conn = self.target_esidxer.conn
        index_name = self.target_esidxer.ES_INDEX_NAME
        index_type = self.target_esidxer.ES_INDEX_TYPE

        #Check if index_type exists
        try:
            conn.get_mapping(index_type, index_name)
        except IndexMissingException:
            return
        return conn.delete_mapping(index_name, index_type)

    def finalize(self):
        conn = self.target_esidxer.conn
        conn.indices.flush()
        conn.indices.refresh()
        self.target_esidxer.optimize()

    def get_id_list(self, step=None):
        return self.target_esidxer.get_id_list(step=step)

    def get_from_id(self, id):
        return self.target_esidxer.get_biothing(id, only_source=True)

    def mget_from_ids(self, ids, step=100000, only_source=True, asiter=True, **kwargs):
        '''ids is an id list. always return a generator'''
        return self.target_esidxer.get_docs(ids, step=step, only_source=only_source, **kwargs)

    def remove_from_ids(self, ids, step=10000):
        self.target_esidxer.delete_docs(ids, step=step)

    def query(self, query=None, verbose=False, step=10000, scroll="10m",
              only_source=True, **kwargs):
        ''' Function that takes a query and returns an iterator to query results. '''
        try:
            return self.target_esidxer.doc_feeder(query=query, verbose=verbose, step=step, scroll=scroll, only_source=only_source, **kwargs)
        except Exception:
            pass

    @classmethod
    def create_from_options(cls, options):
        ''' Function that recreates itself from a DocBackendOptions class.  Probably a needless
        rewrite of __init__... '''
        if not options.es_index or not options.es_host or not options.es_doc_type:
            raise Exception("Cannot create backend class from options, ensure that es_index, es_host, and es_doc_type are set")
        return cls(ESIndexer(index=options.es_index, doc_type=options.es_doc_type, es_host=options.es_host))

class DocBackendOptions(object):
    def __init__(self, cls, es_index=None, es_host=None, es_doc_type=None,
                 mongo_target_db=None, mongo_target_collection=None):
        self.cls = cls
        self.es_index = es_index
        self.es_host = es_host
        self.es_doc_type = es_doc_type
        self.mongo_target_db = mongo_target_db
        self.mongo_target_collection = mongo_target_collection



""" Starting some utils for biothings schema. """

import temp_schema
SCHEMA = temp_schema.SCHEMA

#def generate_avro_schema(schema):

#def generate_web_schema(schema):

#def generate_jsonld_schema(schema):

def generate_es_mappings(schema):
    
    def _traverse(d):
        r = {}
        if '@esmappings' in d:
            r.update({ d['name'] : d['@esmappings'] })
        if 'fields' in d and len(d['fields']) > 0:
            r.setdefault('properties', {})
            for i in d['fields']:
                if isinstance(i, dict):
                    r['properties'].update(dict(_traverse(i)))
        else:
            r.setdefault( d['name'], {} )
            for i in d['type']:
                if isinstance(i, dict) and 'fields' in i:
                    r[ d['name'] ].update(dict(_traverse(i)))
        return list(r.items())
        
    g = _traverse(schema)
    # need to re-think the recursion to prevent this....
    # TODO: remove empties....
    ret = dict([g[0]])
    ret[g[0][0]].update(dict([g[1]]))
    return ret



"""
    Elasticsearch Query Result Transform
"""
from collections import defaultdict

from biothings.utils.common import dotdict
class ResultTransformException(Exception):
    pass

class ESResultTransform(object):
    ''' Class to transform the results of the Elasticsearch query generated prior in the pipeline.
    This contains the functions to extract the final document from the elasticsearch query result
    in `Elasticsearch Query`_.  This also contains the code to flatten a document etc.
    '''

    def __init__(self, web_settings):

        # license appending settings
        self.source_licenses = defaultdict(dict)  # will be populated by web setting
        self.license_transform = web_settings.LICENSE_TRANSFORM

        # mapping transform
        self.field_notes = web_settings.get_field_notes()
        self.excluded_keys = web_settings.AVAILABLE_FIELDS_EXCLUDED

    @classmethod
    def traverse(cls, obj, leaf_node=False):
        """
        Output path-dictionary pairs. For example, input:
        {
            'exac_nontcga': {'af': 0.00001883},
            'gnomad_exome': {'af': {'af': 0.0000119429, 'af_afr': 0.000123077}},
            'snpeff': {'ann': [{'effect': 'intron_variant',
                                'feature_id': 'NM_014672.3'},
                               {'effect': 'intron_variant',
                                'feature_id': 'NM_001256678.1'}]}
        }
        will be translated to a generator:
        (
            ("exac_nontcga", {"af": 0.00001883}),
            ("gnomad_exome.af", {"af": 0.0000119429, "af_afr": 0.000123077}),
            ("gnomad_exome", {"af": {"af": 0.0000119429, "af_afr": 0.000123077}}),
            ("snpeff.ann", {"effect": "intron_variant", "feature_id": "NM_014672.3"}),
            ("snpeff.ann", {"effect": "intron_variant", "feature_id": "NM_001256678.1"}),
            ("snpeff.ann", [{ ... },{ ... }]),
            ("snpeff", {"ann": [{ ... },{ ... }]}),
            ('', {'exac_nontcga': {...}, 'gnomad_exome': {...}, 'snpeff': {...}})
        )
        or when traversing leaf nodes:
        (
            ('exac_nontcga.af', 0.00001883),
            ('gnomad_exome.af.af', 0.0000119429),
            ('gnomad_exome.af.af_afr', 0.000123077),
            ('snpeff.ann.effect', 'intron_variant'),
            ('snpeff.ann.feature_id', 'NM_014672.3'),
            ('snpeff.ann.effect', 'intron_variant'),
            ('snpeff.ann.feature_id', 'NM_001256678.1')
        )
        """
        if isinstance(obj, dict):  # path level increases
            for key in obj:
                for sub_path, val in cls.traverse(obj[key], leaf_node):
                    yield '.'.join((key, sub_path)).strip('.'), val
            if not leaf_node:
                yield '', obj
        elif isinstance(obj, list):  # does not affect path
            for item in obj:
                for sub_path, val in cls.traverse(item, leaf_node):
                    yield sub_path, val
            if not leaf_node or not obj:  # [] count as a leaf node
                yield '', obj
        elif leaf_node:  # including str, int, float, and *None*.
            yield '', obj

    def transform(self, response, options):
        """
        Transform the query result. TODO more
        """
        if not isinstance(options, dotdict):
            options = dotdict(options)
        if isinstance(response, list):
            responses_ = []
            template = options.pop('template', {})
            templates = options.pop('templates', [template]*len(response))
            template_hit = options.pop('template_hit', dict(found=True))
            template_miss = options.pop('template_miss', dict(found=False))
            responses = [self.transform(res, options) for res in response]
            for res_, res in zip(templates, responses):
                if not res.get('hits'):
                    res_.update(template_miss)
                    responses_.append(res_)
                else:
                    for hit in res['hits']:
                        hit_ = dict(res_)
                        hit_.update(template_hit)
                        hit_.update(hit)
                        responses_.append(hit_)
            return list(filter(None, responses_))
        if isinstance(response, dict):
            response.update(response.pop('hits', {}))  # collapse one level
            response.pop('_shards')
            response.pop('timed_out')
            if 'hits' in response:
                for hit in response['hits']:
                    hit.update(hit.pop('_source', {}))  # collapse one level
                    for path, obj in self.traverse(hit):
                        self.transform_hit(path, obj, options)
                        if options.allow_null:
                            self.option_allow_null(path, obj, options.allow_null)
                        if options.always_list:
                            self.option_always_list(path, obj, options.always_list)
                        if options._sorted:
                            self.option_sorted(path, obj)
                    if options.dotfield:
                        self.option_dotfield(hit, options)
            if 'aggregations' in response:
                self.transform_aggregations(response['aggregations'])
                response['facets'] = response.pop('aggregations')
                response['hits'] = response.pop('hits')  # order
            return response
        return {}

    @staticmethod
    def option_allow_null(path, obj, fields):
        """
        The specified fields should be set to None if it does not exist.
        When flattened, the field could be converted to an empty list.
        """
        if isinstance(obj, dict):
            for field in fields or []:
                if field.startswith(path):
                    key = field[len(path):].lstrip('.')
                    if '.' not in key and key not in obj:
                        obj[key] = None

    @staticmethod
    def option_always_list(path, obj, fields):
        """
        The specified fields, if exist, should be set to a list type.
        None converts to an emtpy list [] instead of [None].
        """
        if isinstance(obj, dict):
            for field in fields:
                if field.startswith(path):
                    key = field[len(path):].lstrip('.')
                    if key in obj and not isinstance(obj[key], list):
                        obj[key] = [obj[key]] if obj[key] is not None else []

    @staticmethod
    def option_sorted(_, obj):
        """
        Sort a container in-place.
        """
        try:
            if isinstance(obj, list):
                obj.sort()  # TODO list of object might is a problem here
            elif isinstance(obj, dict):
                sorted_items = sorted(obj.items())
                obj.clear()
                obj.update(sorted_items)
        except Exception:
            pass  # TODO logging

    @classmethod
    def option_dotfield(cls, dic, options):
        """
        Flatten a dictionary.
        #TODO examples
        """
        hit_ = defaultdict(list)
        for path, value in cls.traverse(dic, leaf_node=True):
            hit_[path].append(value)
        for key, lst in hit_.items():
            if len(lst) == 1 and key not in (options.always_list or []):
                hit_[key] = lst[0]
            else:  # multi-element list
                hit_[key] = list(filter(None, lst))
                if options._sorted:
                    cls.option_sorted(key, hit_[key])
        dic.clear()
        dic.update(hit_)

    def transform_hit(self, path, doc, options):
        """
        By default add licenses

        If a source has a license url in its metadata,
        Add "_license" key to the corresponding fields.
        Support dot field representation field alias.

        If we have the following settings in web_config.py

        LICENSE_TRANSFORM = {
            "exac_nontcga": "exac",
            "snpeff.ann": "snpeff"
        },

        Then GET /v1/variant/chr6:g.38906659G>A should look like:
        {
            "exac": {
                "_license": "http://bit.ly/2H9c4hg",
                "af": 0.00002471},
            "exac_nontcga": {
                "_license": "http://bit.ly/2H9c4hg",         <---
                "af": 0.00001883}, ...
        }
        And GET /v1/variant/chr14:g.35731936G>C could look like:
        {
            "snpeff": {
                "_license": "http://bit.ly/2suyRKt",
                "ann": [{"_license": "http://bit.ly/2suyRKt", <---
                        "effect": "intron_variant",
                        "feature_id": "NM_014672.3", ...},
                        {"_license": "http://bit.ly/2suyRKt", <---
                        "effect": "intron_variant",
                        "feature_id": "NM_001256678.1", ...}, ...]
            }, ...
        }

        The arrow marked fields would not exist without the setting lines.
        """
        if path == '':
            doc.pop('_index')
            doc.pop('_type', None)    # not available by default on es7
            doc.pop('sort', None)     # added when using sort
            doc.pop('_node', None)    # added when using explain
            doc.pop('_shard', None)   # added when using explain

        licenses = self.source_licenses[options.biothing_type]
        if path in self.license_transform:
            path = self.license_transform[path]
        if path in licenses and isinstance(doc, dict):
            doc['_license'] = licenses[path]

    def transform_aggregations(self, res):
        for facet in res:
            res[facet]['_type'] = 'terms'
            res[facet]['terms'] = res[facet].pop('buckets')
            res[facet]['other'] = res[facet].pop('sum_other_doc_count')
            res[facet]['missing'] = res[facet].pop('doc_count_error_upper_bound')
            count = 0
            for term in res[facet]['terms']:
                term['count'] = term.pop('doc_count')
                count += term['count']
                term['term'] = term.pop('key')
                for agg_k in list(term.keys()):
                    if agg_k not in ['count', 'term']:
                        term.update(self.transform_aggregations({agg_k: term[agg_k]}))
            res[facet]['total'] = count
        return res

    def transform_mapping(self, mapping, prefix, search):
        """
        Transform Elasticsearch mapping definition to
        user-friendly field definitions metadata result
        """
        assert isinstance(mapping, dict)
        assert isinstance(prefix, str) or prefix is None
        assert isinstance(search, str) or search is None

        result = {}
        todo = list(mapping.items())
        todo.reverse()

        while todo:
            key, dic = todo.pop()
            dic = dict(dic)
            dic.pop('dynamic', None)
            dic.pop('normalizer', None)

            if key in self.field_notes:
                result['notes'] = self.field_notes[key]

            if 'copy_to' in dic:
                if 'all' in dic['copy_to']:
                    dic['searched_by_default'] = True
                del dic['copy_to']

            if 'index' not in dic:
                if 'enabled' in dic:
                    dic['index'] = dic.pop('enabled')
                else:  # true by default
                    dic['index'] = True

            if 'properties' in dic:
                dic['type'] = 'object'
                subs = (('.'.join((key, k)), v) for k, v in dic['properties'].items())
                todo.extend(reversed(list(subs)))
                del dic['properties']

            if all((not self.excluded_keys or key not in self.excluded_keys,
                    not prefix or key.startswith(prefix),
                    not search or search in key)):
                result[key] = dict(sorted(dic.items()))

        return result




def run_once():
    """
    should_run_task_1 = run_once()
    print(should_run_task_1()) -> True
    print(should_run_task_1()) -> False
    print(should_run_task_1()) -> False
    print(should_run_task_1()) -> False

    should_run_task_2 = run_once()
    print(should_run_task_2('2a')) -> True
    print(should_run_task_2('2b')) -> True
    print(should_run_task_2('2a')) -> False
    print(should_run_task_2('2b')) -> False
    ...
    """

    has_run = set()

    def should_run(identifier=None):

        if identifier in has_run:
            return False

        has_run.add(identifier)
        return True

    return should_run
